(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)
(*  Soubor: VESA.PAS                                                       *)
(*  Obsah: kompletni a sobestacna jednotka pro VESA grafiku                *)
(*  Autori puvodni jednotky SVGA 4: Asp / VR group, 1996-99                *)
(*                     Virtual Research independent group production, 1999 *)
(*  Prevzal, upravil, opravil, rozsiril, vylepsil: Mircosoft               *)
(*  Posledni uprava: 19.10.2006                                            *)
(*  Pro kompilaci: nic                                                     *)
(*  Pro spusteni: pokud chcete zobrazovat texty, potrebujete nejaky soubor *)
(*                s fontem ve formatu *.FNT (vytvareji se mym programem    *)
(*                FNTEDIT, neni to zadny oficialni format), jinak nic.     *)
(*  Upozorneni: tyto zdrojove kody pouzivate na vlastni nebezpeci          *)
(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)
unit VESA;
{$G+,N+,E-,S-,Q-,R-,V-,B-,I-}
{...$D-,L-}

{...$define Blbuvzdornost} {nechte definovat, pokud chcete zapnout kontrolu
poradi parametru pro _Hline, Vline, Rectangle, Bar, Box, Maskedbar a
Maskedbox. Pri vypnute kontrole se sice nepatrne zrychli, ale zase si musite
davat pozor, abyste souradnice neprohodili (jinak se cela banka zaplni
bordelem). __Xorrectangle a __Hline se automaticky kontroluji vzdy.}

interface

type _pole8bytu=array[0..7]of byte; {na vzorek vyplne pro procedury Maskedbar a spol.}

     _FontArray=array[0..0]of byte;
     {Pomocna datova struktura pro ukladani rastrovych fontu. Delka pole je
     dynamicka podle vysky znaku. Sam o sobe je tento typ nepouzitelny -
     - potrebujeme znat vysku pisma, jinak se v nem nic nenajde.
     Technicke detaily: kazde pismeno je tvoreno nekolika byty, kazdy
     predstavuje jeden radek (kde je jednicka, tam se pixel vykresli, kde je
     nula, tam ne). Sirka znaku je pevna - 8 pixelu (tj. 8 bitu v 1 bytu),
     vyska libovolna. Znakova sada zabere v pameti 256 krat vyska pisma bytu
     (pro pismo 8x8 pixelu to tedy dela presne 2 kB).}
     _UkNaFont=^_fontarray; {dalsi pomocny typ}

     _Font = record {tohle je k praktickemu pouziti}
             fVyska:byte;{vyska pisma v radcich (bytech)}
             fStdX,fStdY,fStdRoztec:byte;{doporucene hodnoty ulozene v souboru s fontem}
             fData:_uknafont;
             end;

var _MaxX,_MaxY:word; {pravy dolni roh obrazovky (nastavi se automaticky pri spusteni grafiky)}
    _MinClipX,_MinClipY,_MaxClipX,_MaxClipY:integer; {orezavaci okno}
    _Mode:word; {aktualni graficky rezim}

procedure _SetMode(mode:word);
{Nastavi prislusny graficky (nebo textovy) rezim. Moznosti jsou:}
const _Textak=0; {obycejny 80x25}
      _640x350=$11C;
      _640x400=$100;
      _640x480=$101;
      _800x600=$103;
      _1024x768=$105;

procedure _SwitchBank(bank:word);
{Prepne do zadane banky (cislovani zacina od 0). Obvykle neni potreba ji volat
rucne, ale nekdy se hodi, napr. v kombinaci s tou nasledujici do procedur
bezicich na pozadi.}
function _GetBank:word;
{vrati cislo aktualne nastavene banky}

{kreslici procedury bez orezavani (pozor, nesmi vycuhovat z obrazovky!):}

procedure _PutPixel(x,y:word;color:byte);
{nakresli pixel}
procedure _PutTPixel(x,y:word;color:byte);
{kdyz je pozadovana barva pixelu rovna "pruhledne" barve (0), nic se nekresli,
jinak normalni putpixel}
function _GetPixel(x,y:word):byte;
{vrati barvu pixelu na danych souradnicich}
procedure _Line(x1,y1,x2,y2:word; barva:byte);
{cara}
procedure _HLine(x1,x2,y:word; color:word);
{velmi rychla (32b) vodorovna cara. Barva je word jenom kvuli tomu, aby se
s ni pak lip pracovalo, hodnoty davejte normalne 0..255.}
procedure _VLine(x,y1,y2:word; color:byte);
{svisla cara}
procedure _MaskedLine(x1,y1,x2,y2:integer; barva:byte; maska:word);
{Maskovana cara. Maska je podobna jako pri nastavovani SetLineStyle v BGI.
Vykresli se jen ty bity, ve kterych je 1. Nejvyssi bit masky se promitne do
bodu x1;y1. }
procedure _Ellipse(xc,yc,rx,ry:word; color:byte);
{prazdna elipsa.
 xc, yc - souradnice stredu
 rx, ry - polomery ve smeru x a y}
procedure _FilledEllipse(xc,yc,rx,ry:word; color:byte);
{vyplnena elipsa}
procedure _Rectangle(x1,y1,x2,y2:word; color:byte);
{prazdny obdelnik
 x1, y1 - levy horni roh
 x2, y2 - pravy dolni roh}
procedure _Bar(x1,y1,x2,y2:word; barva:word);
{vyplneny obdelnik, dost rychly (32b). K barve stejny komentar jako u HLine}
procedure _MaskedBar(x1,y1,x2,y2,barva:word; var vzorek:_pole8bytu);
{obdelnik s dvojbarevnou vyplni, podobny tomu z BGI se vzorkem nastavenym
procedurou Setfillpattern. Navazovani vzorku pri kresleni nekolika obdelniku
pres sebe je zaruceno, rychlost je pomerne znacna (64b).
 barva - v hornim bytu je barva pozadi, v dolnim barva popredi
 vzorek - pole osmi bytu. Kazdy byte je jeden radek, bity s hodnotou 1 se
          vykresli barvou popredi a bity 0 barvou pozadi. Nejvyssi bit bude
          nejvic vlevo, radek (byte) s indexem 0 bude nejvic nahore. Berte to
          jako ctverec 8x8 bitu.
Nejake preddefinovane vzorky:}
const _RidkeSikmeSrafy:_pole8bytu=($01,$02,$04,$08,$10,$20,$40,$80);{/}
      _HusteSikmeSrafy:_pole8bytu=($11,$22,$44,$88,$11,$22,$44,$88);
      _RidkeKosmeSrafy:_pole8bytu=($80,$40,$20,$10,$08,$04,$02,$01);{\  (nazev je oficialni z heraldiky :-) )}
      _HusteKosmeSrafy:_pole8bytu=($88,$44,$22,$11,$88,$44,$22,$11);
      _SikmePruhy:     _pole8bytu=($78,$F0,$E1,$C3,$87,$0F,$1E,$3C);
      _KosmePruhy:     _pole8bytu=($3C,$1E,$0F,$87,$C3,$E1,$F0,$78);
      _KolmaMrizka:    _pole8bytu=($22,$22,$FF,$22,$22,$22,$FF,$22);{+}
      _SikmaMrizka:    _pole8bytu=($11,$0A,$04,$0A,$11,$A0,$40,$A0);{x}
      _VelmiRidkeTecky:_pole8bytu=($00,$02,$00,$00,$00,$20,$00,$00);
      _RidkeTecky:     _pole8bytu=($00,$22,$00,$00,$00,$22,$00,$00);
      _StredniTecky:   _pole8bytu=($00,$22,$00,$88,$00,$22,$00,$88);
      _HusteTecky:     _pole8bytu=($00,$AA,$00,$AA,$00,$AA,$00,$AA);
      _VelmiHusteTecky:_pole8bytu=($55,$AA,$55,$AA,$55,$AA,$55,$AA);

procedure _Box(x1,y1,x2,y2:word; LHbarva,PDbarva,Vbarva:word);
{obdelnik s obrysem (jako napr. tlacitko). Dost rychly (32b).
 LHbarva - barva leveho a horniho okraje
 PDbarva - barva praveho a dolniho okraje
 Vbarva - barva vyplne. Pokud je 0, vypln se nekresli (vnitrek tedy nebude
          v barve 0, ale zustane prazdny!).}
procedure _MaskedBox(x1,y1,x2,y2:word; lhbarva,pdbarva,vbarva:word; var vzorek:_pole8bytu);
{vysledek je podobny jako u _Box, ale vnitrek se kresli procedurou _MaskedBar.
Kdyz je vbarva=0, vnitrek se nekresli.}
procedure _Triangle(x1,y1,x2,y2,x3,y3:word;color:byte);
{prazdny trojuhelnik (slozeny ze tri car)}
procedure _FilledTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
{Vyplneny trojuhelnik. Je nutne zapnout Blbuvzdornost, protoze je tam par
nedostatku, takze parametry se caram predavaji obracene nez by bylo potreba.}
procedure _FilledPolygon(var VertexArray; NumOfVertexes,color:byte);
{vyplneny mnohouhelnik. NumOfVertexes je pocet vrcholu, VertexArray je pole
souradnic (dvojic cisel typu integer). Nutnou a postacujici podminkou pro
spravne vykresleni je, aby kazda (pomyslna) usecka vedena z prvniho vrcholu do
libovolneho jineho lezela cela uvnitr obrysu mnohouhelniku. Postacujici
podminkou tedy je, aby byl mnohouhelnik konvexni.}
procedure _Fill(color:word);
{vyplni celou obrazovku danou barvou (32b); funguje jen pri normalnim
zobrazeni, ne pri strankovani}
procedure _Fill2(color:word);
{podobna, ale: 1) funguje pri strankovani, protoze 2) je to jenom _bar pres
celou obrazovku}
procedure _XorPixel(x,y:word; color:byte);
procedure _XorLine(x1,y1,x2,y2:word; barva:byte);
procedure _XorHLine(x1,x2,y:word; color:word);
procedure _XorVLine(x,y1,y2:word; color:byte);
procedure _XorMaskedLine(x1,y1,x2,y2:integer; barva:byte; maska:word);
procedure _XorEllipse(xc,yc,rx,ry:word; color:byte);
procedure _XorRectangle(x1,y1,x2,y2:word; color:byte);
procedure _XorTriangle(x1,y1,x2,y2,x3,y3:word; color:byte);
{podobne vyse uvedenym, ale jejich barva se xoruje s pozadim}

{getimage, putimage atd.:}
procedure _GetImage(x1,y1,x2,y2:word;clip:pointer);
{nacte vyrez obrazovky - podobne jako BGI Getimage. Nacitana oblast by nemela
zasahovat mimo obrazovku (pokud ano, nactou se kraviny). Potrebne mnozstvi
pameti pro obrazek se spocita jako sirka*vyska obrazku (1 pixel = 1B,
informace o rozmerech se neukladaji), obrazova data jsou v pameti usporadana
po radcich zleva doprava a odshora dolu.}
procedure _PutImage(x1,y1:integer;width,height:word;clip:pointer);
{zobrazi nacteny obrazek. Pozor, ze se musi zadat prislusna sirka a vyska.
Muze zasahovat mimo obrazovku, vykresli se jen viditelna cast.}
procedure _PutTImage(x1,y1:integer;width,height:word;clip:pointer);
{podobny jako _Putimage, ale pixely s barvou 0 se nevykresli (pruhlednost)}
procedure _PutRTImage(x1,y1:integer;width,height:word;clip:pointer);
{podobny jako _PutTImage, ale kresli obrazek zrcadlove prevraceny
(leva <-> prava strana)}
{vsechny _put... se orezavaji podle nastavenych okraju, jako ty nasledujici:}

{kreslici procedury s orezavanim (oriznou se tak, aby se zobrazila jenom ta
cast, ktera je uvnitr nastavenych okraju (_minclipx, _minclipy, _maxclipx,
_maxclipy). Parametry a pouziti stejne jako u tech bez orezavani.}
procedure __PutPixel(x,y:integer;color:byte);
procedure __XorPixel(x,y:integer;color:byte);
procedure __PutTPixel(x,y:integer;color:byte);
function  __GetPixel(x,y:integer):byte;
procedure __Line(x1,y1,x2,y2:integer;color:byte);
procedure __XorLine(x1,y1,x2,y2:integer;color:byte);
procedure __MaskedLine(x1,y1,x2,y2:integer;color:byte;mask:word);
procedure __XorMaskedLine(x1,y1,x2,y2:integer;color:byte;mask:word);
procedure __Ellipse(xc,yc:integer;rx,ry:word;color:byte);
procedure __XorEllipse(xc,yc:integer;rx,ry:word;color:byte);
procedure __FilledEllipse(xc,yc:integer;rx,ry:word;color:byte);
procedure __Rectangle(x1,y1,x2,y2:word;color:byte);
procedure __XorRectangle(x1,y1,x2,y2:word;color:byte);
procedure __Bar(x1,y1,x2,y2:integer;vbarva:byte);
procedure __MaskedBar(x1,y1,x2,y2:integer;barva:word;var vzorek:_pole8bytu);
procedure __Box(x1,y1,x2,y2:integer;lhbarva,pdbarva,vbarva:byte);
procedure __MaskedBox(x1,y1,x2,y2:word;lhbarva,pdbarva:byte;vbarva:word;var vzorek:_pole8bytu);
procedure __HLine(x1,x2,y:integer;color:byte);
procedure __XorHLine(x1,x2,y:integer;color:byte);
procedure __VLine(x,y1,y2:integer;color:byte);
procedure __XorVLine(x,y1,y2:integer;color:byte);
procedure __Triangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
procedure __XorTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
procedure __FilledTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
procedure __FilledPolygon(var VertexArray;NumOfVertexes,color:byte);
procedure __Fill(color:byte);

procedure __FloodFill(ffx,ffy:integer; barva:byte);
{Funguje jako plechovka v Malovani: barva se "rozlije" od bodu [ffx,ffy] do
vsech stran a vyplni celou jednobarevnou oblast uzavrenou bud okrajem jine
barvy nebo okrajem orezavaciho okna.
Procedura potrebuje pomerne dost pameti, ale je udelana tak, ze i kdyz pamet
dojde, je vetsinou schopna zadanou oblast vyplnit, ale vychozi bod musi byt
zadan nekde priblizne uprostred oblasti, ne treba u horniho okraje.
Jo, a abych nezapomnel... je nechutne pomala.}


{pro psani textu:}
procedure _NactiFont(var FN:_font; var Soubor:file);
{Nacte ze souboru font.
FN - promenna typu _Font. Kdyz je po skonceni procedury FN.fdata=nil, znamena
     to, ze se nekde stala chyba a nacteni se nepovedlo. Nelze pouzivat na jiz
     nactene fonty!!! Ty by zustaly ztracene v pameti.
Soubor - beztypovy soubor otevreny pro cteni s velikosti bloku 1 byte
         (tj. reset(soubor,1);), kurzor v nem musi byt na zacatku fontu.
         Procedura soubor nezavre (to vsechno proto, aby sel font zabudovat
         do nejakeho vetsiho datoveho souboru).
Format souboru s fontem je popsan na konci teto jednotky.}
procedure _NactiFont2(var FN:_font; JmenoSouboru:string);
{totez, ale zadava se jmeno souboru, ktery si procedura otevre a pak zase
zavre. Jednodussi na pouziti, ale lze pouzit jen pro soubory obsahujici
pouze samotny font.}
function _UlozFont(FN:_font;soubor:string):boolean;
{ulozi font do souboru (jmeno piste i s koncovkou), vraci true, pokud se to
povede. Pro pouziti zrejme v nejakem programu, ve kterem se ten font nejdriv
vytvori.}
procedure _ZrusFont(var FN:_font);
{vymaze font z pameti}
procedure _NastavFont(FN:_font; LoadDefault:boolean);
{nastavi font pro psani procedurou _Print (viz dale). Promennou F je potreba
zachovat, protoze se na ni pouze napichnou prislusne ukazatele, font se nikam
fyzicky nekopiruje.
 LoadDefault - true => Nastavi se doporucene parametry textu (velikost a
                       roztec), ktere byly s fontem ulozeny v souboru.
               false => Nastaveni textu se nezmeni.}
procedure _NastavText(velikostX,velikostY,roztec:byte);
{nastavi styl textu pro proceduru _Print. Kdyz je jakykoli parametr roven 0,
zustava nastavena jeho minula hodnota (tj. nemeni se).
 velikostX, velikostY - velikost pisma, technicky vzato rozmer jednoho pixelu,
                        pro vodorovny a svisly smer.
 roztec - roztec dvou sousednich znaku. Normalne je rovna sirce znaku;
          kdyz menite velikost pisma, nezapomente zadat roztec rovnou
          8*velikostX, jinak bude pismo zdrcle dohromady (coz se nekdy hodi,
          proto tenhle system).}
procedure _PushText;
{ulozi vsechny parametry textu (rozmery, roztec i ukazatel na font)
do dynamickeho LIFO zasobniku}
procedure _PopText;
{znovu nastavi posledne ulozene parametry textu a odstrani je ze zasobniku.
Pouzito napr. v ruznych dialogovych oknech, ktere si nastavi vlastni styl
pisma a pak zase obnovi ten, ktery byl nastaven puvodne.}
procedure _LodText;
{Znovu nastavi posledne ulozene parametry textu, ale neodstrani je ze
zasobniku. Obcas se taky hodi.}
procedure _Print(ix,iy:integer; barva:byte; textik:string);
{napise na obrazovku text (podobne jeko Outtextxy). Orezava podle orezavaciho
okna. Kdyby kus nejakeho pismena vycuhoval vlevo nebo vpravo, cele to pismeno
se nezobrazi; kdyby vycuhovalo nahore nebo dole, zobrazi se jen viditelna
cast.
Do zobrazovaneho retezce se daji vkladat specialni ridici kody. Vzdycky
zacinaji znaky '%%', pak nasleduje jeden ridici znak a pak urcity pocet
dalsich znaku - parametru. Jsou delany maximalne blbuvzdorne, tj. pri jakekoli
chybe se proste zobrazi jako normalni text. Zatim funguji tyto:
 %%b### - nastavi barvu textu. ### je triciferne cislo barvy (000..255).
          Je nutne ho vzdy doplnit nulami na zacatku na tri cifry.
 %%z### - misto tohoto kodu zobrazi znak s ordinalnim (ASCII) cislem ###
          (opet nutne triciferne). Vhodne napr. pokud chcete zobrazit znaky ze
          zacatku ASCII tabulky, ktere ale v textovem souboru zpusobi chaos
          (existuje napr. znak "konec souboru" a zkuste pak z takoveho souboru
          precist neco, co je za nim...).
 Pokud z nejakeho duvodu potrebujete ridici sekvence vyradit z provozu, zmente
hodnotu teto promenne...}
const _RidiciKody:byte=1;
{...temito hodnotami:}
      _pouzivat=1;{pouzivat, ridit se podle nich}
      _vypisovat=2;{nepouzivat, zobrazit jako prosty text}
      _preskakovat=3;{nepouzivat, preskocit a nezobrazovat je}
function _SirkaTextu(retezec:string):word;
{vraci sirku zadaneho retezce v pixelech (obdoba Textwidth)}
function _SirkaTextu2(PocetZnaku:byte):word;
{vraci totez, ale misto retezce zadate jen pocet znaku (font je
neproporcionalni - vsechny znaky jsou stejne siroke)}
function _VyskaTextu:word;
{vraci vysku jakehokoli textu v pixelech (obdoba Textheight, ale nezalezi na
tom, co se pise, protoze vsechny znaky fontu jsou stejne vysoke)}

{Obcas je v programech potreba primy pristup k nastavenym parametrum textu,
takze tady jsou:}
var __velikostX,__velikostY:byte; {velikost zobrazovaneho pisma, zvlast pro vodorovny a svisly smer}
    __roztec:byte; {roztec zobrazovanych znaku (od leveho kraje jednoho po levy kraj druheho)}


{ostatni:}
procedure _SetWindow(minx,miny,maxx,maxy:integer);
{nastavi orezavaci okno; pokud by nejakou stranou zasahovalo mimo obrazovku,
nastavi se ta strana na okraj obrazovky}
procedure _SetDefaultWindow;
{nastavi orezavaci okno na celou obrazovku}
procedure _VSync;
{ceka na synchronizacni impuls obrazovky}
procedure _VSyncStart;
{ceka na zacatek synchronizacniho impulsu (tj. jestli impuls pri zavolani
procedury zrovna probiha, pocka, az skonci a zacne podruhe; tim se ziska
trochu vic casu, ale taky se nekdy trochu dele ceka)}

{pro manipulaci s virtualni obrazovkou:}
procedure _SetDisplayOrigin(x,y:word);
{nastavi polohu leveho horniho rohu monitoru nad virtualni obrazovkou.
Normalni hodnota je 0,0.}
procedure _SetAP(kterou:byte);
procedure _SetVP(kterou:byte);
{SetActivePage (nastaveni stranky, na kterou se bude kreslit) a SetVisualPage
(nastaveni stranky, ktera bude videt na obrazovce). Parametr Kterou je od 0
do X, kde X je jen tak velke, aby vystacila videopamet.}
function _PujdeStrankovani:boolean;
{zjisti, jestli je pri aktualnim rozliseni dost videopameti aspon na jednu
virtualni obrazovku}
procedure _NastavStrankovani;
{Prepina mezi normalnim a strankovacim rezimem. Viditelna stranka zustava vzdy
na miste a aktivni se prepne - bud na tu druhou stranku nez je viditelna,
pokud strankovani zapiname, nebo na stejnou jako viditelna, pokud ho vypiname.
Je treba hlidat si, jestli mame strankovani zapnute nebo vypnute, aby nevznikl
zmatek. V pripade, ze presto vznikne, pomuze nasledujici procedura:}
procedure _ResetStrankovani;
{Natvrdo nastavi aktivni i viditelnou stranku na 0.
Mimochodem, po vypnuti strankovani obvykle neni jasne, ktera stranka je
zrovna nastavena. Vetsine procedur to nevadi, ale vyjimkou je _Fill, ktera
pracuje jen se strankou 0. Takze bud pouzivejte jeji o malinko pomalejsi
modifikaci _Fill2 nebo resetujte stranky.}
procedure _Swap;
{Prohodi stranku pro zobrazeni a pro kresleni; pri zapnutem strankovani je
nutne ji volat pokazde, kdyz chceme nechat zobrazit to, co jsme nakreslili.
Ma zabudovane cekani na navrat paprsku (synchronizacni impuls obrazovky),
takze obraz neblika.}

{...$define PovolShowbanks}
{$ifdef povolshowbanks}
procedure _ShowBanks;
{Pro testovaci ucely. Kazdou banku na obrazovce vyplni jinou barvou, aby bylo
videt, kde presne zacinaji a konci.}
{$endif}

implementation
uses errmsg;

var _VelikostVideopameti:word; {velikost videopameti v 64 kB blocich}
    Granularity:word;{granularita znamena, o kolik kB dal zacina nasledujici
                      banka. Tohle uz je prepocitana hodnota, kterou staci
                      vynasobit cislo banky spocitane beznym zpusobem.}
    CurBank:word;         {aktualni banka}
    __SwitchBank:pointer; {adresa procedury pro prepinani bank pod VESA}
    __LHy:word;           {pricita se k yove souradnici vseho, co se kresli
                           (kvuli strankovani)}

    __Font:_uknafont; {aktualni font...}
    __Vyska:byte; {...a jeho vyska}

procedure _initVESA;{detekce, inicializace atd.}
type
T_VESAInfo = record{informace o VESA hardwaru}
             Signature:array[0..3] of char; {'VESA' kdyz je vsechno v poradku}
             Version:word;       {cislo verze, obcas se hodi}
             OEMId:pointer;      {ukazatel na tabulku informaci o vyrobci, celkem k nicemu}
             Flags:longint;      {podrobnosti o hardwaru}
             ModeList:pointer;   {ukazatel na tabulku dostupnych videorezimu}
             VideoMemory:word;   {mnozstvi videopameti v 64 kB blocich}
             Reserved:array[1..236] of byte;
             pojistka:array[1..256] of byte; {verze 2.0 potrebuje 512 B buffer, tak kdyby nahodou...}
             end;
var _VESAInfo:^T_VESAInfo;
    vysledek:word;
Begin
new(_vesainfo);
asm
mov AX,$4F00         {kod sluzby - "zjisti informace o VESA rozhrani"}
les DI,_vesainfo     {adresa zasobniku pripravena}
int $10              {a makej...}
mov vysledek,AX      {jak jsme dopadli?}
end;
if (vysledek<>$004F)or((_VESAInfo^.Signature<>'VESA')and(_VESAInfo^.Signature<>'VBE2')) then
  chyba('Vase graficka karta bohuzel nema VESA rozhrani, ktere je pro spusteni tohoto programu potreba.',0);
_VelikostVideopameti:=_vesainfo^.videomemory;
dispose(_vesainfo);
_Mode:=0;
End;{_initVESA}

procedure _SwitchBank(bank:word); assembler;
{procedura pro prepinani bank s osetrenim granularity - abych ji nemusel
rozepisovat uplne vsude}
Asm
mov AX,bank            {AX:=pozadovana banka}
mov CurBank,AX         {ulozeni banky do globalni promenne}
mul Granularity        {vynasobeni AX granularitou - ted je tam skutecne cislo banky}
mov DX,AX              {do DX s nim (__switchbank ho chce prave tam)}
mov AX,$4F05           {kod sluzby - "ovladani pristupu k obrazovce"}
xor BX,BX              {jeste podkod - "nastav banku"}
call __switchbank      {tak do toho!}
End;{_switchbank}

function _getbank:word; assembler;
Asm
mov AX,$4F05           {kod sluzby - "ovladani pristupu k obrazovce"}
mov BX,$0100           {podkod - "zjisti banku"}
call __switchbank      {do toho!}
mov AX,DX              {v DX se nam vratilo cislo banky}
xor DX,DX              {DX bude vyssi word delence, tak musi byt 0}
div granularity        {vydelime AX granularitou, v DX zustane zbytek, ale ten nas nezajima}
End;{_switchbank}      {to je vsechno, v AX se vrati vysledek}

(* Nasleduje nedokonceny pokus o spusteni grafiky s dukladnejsimi
bezpecnostnimi kontrolami (jestli vybrany rezim opravdu odpovida rozlisenim a
barevnou hloubkou). Kazdopadne jsem tam neco zvoral, takze to nefunguje.
Puvodni jednodussi a hlavne fungujici varianta nastesti existuje nadale :-).

procedure _SetMode(mode:word);
type ModeInfo = record
                Atributy:word;
                AtributyOknaA:byte;
                AtributyOknaB:byte;
                Granularita:word;{[kB]}
                VelikostOkna:word;{[kB]}
                SegmentOknaA:word;{obvykle $A000}
                SegmentOknaB:word;
                BankovaciProcedura:pointer;
                BytuNaRadek:word;
                {pokud (atributy and 2)=2, jsou nastaveny nasledujici hodnoty:}
                RozliseniX:word;
                RozliseniY:word;
                SirkaZnaku,VyskaZnaku:byte;
                PocetBitovychRovin:byte;
                BituNaPixel:byte;
                PocetBank:byte;
                PametovyModel:byte;
                VelikostBanky:byte;{[kB]}
                PocetStranek:byte;{od verze 1.1}
                Zbytek:array[1..482] of byte;{doplnek do nutnych 512 B}
                end;
     ModeInfoPtr=^modeinfo;
var i:word;
    Info:modeinfoptr;
{}procedure OkoukniRezim;
{}var vysledek:word;
{}Begin
{}asm
{}mov AX,$4F01    {sluzba "informace o rezimu"}
{}mov CX,_mode
{}les DI,info
{}int $10
{}mov vysledek,AX
{}end;
{}if (vysledek<>$004F) then chyba('Vase graficka karta neni plne kompatibilni s VESA standardem, '+
{}                                'nefunguje jedna z potrebnych zakladnich funkci.',0);
{}End;{okouknirezim}
Begin{_setmode}
_Mode:=mode;
case _mode of _Textak:begin
                      asm
                      mov ax,3
                      int $10
                      end;
                      Exit;
                      end;
              _640x350:begin
                       _MaxX:=639;
                       _MaxY:=349;
                       end;
              _640x400:begin
                       _MaxX:=639;
                       _MaxY:=399;
                       end;
              _640x480:begin
                       _MaxX:=639;
                       _MaxY:=479;
                       end;
              _800x600:begin
                       _MaxX:=799;
                       _MaxY:=599;
                       end;
              _1024x768:begin
                        _MaxX:=1023;
                        _MaxY:=767;
                        end;
              end;
new(Info);
okouknirezim;
i:=0;
while ((((info^.atributy and 2)<>0)and((info^.RozliseniX<>_maxx+1)or(info^.RozliseniY<>_maxy+1)))
       or(((info^.atributy and 2)=0)and(_mode<>mode)))and(_polerezimu^[i]<>$FFFF)
       {tj. dokud nemame ten spravny rezim a neprosli jsme vsechny dostupne}
       {jeje... jeste jsem ani nezkontroloval, jestli je to 256barevny rezim}
   do begin
      _mode:=_polerezimu^[i];{zkusime nejaky jiny dostupny rezim}
      okouknirezim;{zjistime si o nem informace}
      inc(i);
      end;{while}
if (_polerezimu^[i]=$FFFF)or((info^.atributy and 1)=0) then chyba('Pozadovany graficky rezim neni vasi kartou podporovan.',0);
with info^ do
 begin
 if SegmentOknaA<>$A000 then chyba('Jiny segment obrazovky nez $A000!',0);
 {snad vsechny kontrolni postupy uz jsou vycerpany, ted vzhuru na ty praktictejsi veci:}
 if granularita=0 then granularity:=1{aby se na dalsim radku nedelilo nulou}
                  else granularity:=64 div granularita;{prepocitani granularity na pouzitelnou hodnotu (pro nasobeni)}
          {(granularita je polozka infa a granularity je globalni promenna)}
 __switchbank:=bankovaciprocedura;
 end;{with info^}
dispose(info);
asm{nastaveni rezimu}
mov AX,$4F02     {sluzba "nastav rezim"}
mov BX,mode
int $10
mov i,AX
end;
if (i<>$004F) then chyba('Nepovedlo se spustit graficky rezim.',0);
_setdefaultwindow;
CurBank:=0;{po inicializaci je nastavena banka 0}
End;{_setmode}*)

procedure _SetMode(mode:word);
type ModeInfo = record{informace o rezimu}
                Atributy:word;
                AtributyOknaA:byte;
                AtributyOknaB:byte;
                Granularita:word;{[kB]}
                VelikostOkna:word;{[kB]}
                SegmentOknaA:word;{obvykle $A000}
                SegmentOknaB:word;
                BankovaciProcedura:pointer;{dulezita procedura na prepinani bank}
                BytuNaRadek:word;
                {pokud (atributy and 2)=2, jsou nastaveny nasledujici hodnoty:}
                RozliseniX:word;
                RozliseniY:word;
                SirkaZnaku,VyskaZnaku:byte;
                PocetBitovychRovin:byte;
                BituNaPixel:byte;
                PocetBank:byte;
                PametovyModel:byte;
                VelikostBanky:byte;{[kB]}
                PocetStranek:byte;{od verze 1.1}
                Zbytek:array[1..482] of byte;{doplnek do nutnych 512 B}
                end;
     ModeInfoPtr=^modeinfo;
var vysledek:word;
    Info:modeinfoptr;
Begin
_Mode:=mode;
case _mode of _Textak:begin
                      asm
                      mov ax,3 {kod obycejneho textoveho rezimu 80x25, 16 barev}
                      int $10
                      end;
                      Exit;
                      end;
              _640x350:begin
                       _MaxX:=639;
                       _MaxY:=349;
                       end;
              _640x400:begin
                       _MaxX:=639;
                       _MaxY:=399;
                       end;
              _640x480:begin
                       _MaxX:=639;
                       _MaxY:=479;
                       end;
              _800x600:begin
                       _MaxX:=799;
                       _MaxY:=599;
                       end;
              _1024x768:begin
                        _MaxX:=1023;
                        _MaxY:=767;
                        end;
              end;
new(Info);
asm
mov AX,$4F01    {sluzba "zjisti informace o rezimu"}
mov CX,_mode    {o kterem rezimu}
les DI,info     {pripravime zasobnik}
int $10
end;
with info^ do begin
              if granularita=0 then granularity:=1{aby se na dalsim radku nedelilo nulou}
                               else granularity:=64 div granularita;{prepocitani granularity na pouzitelnou hodnotu}
     {(pozn.: granularita je polozka infa a granularity je globalni promenna)}
              __switchbank:=bankovaciprocedura;
              end;
dispose(info);
asm
mov AX,$4F02     {sluzba "nastav rezim"}
mov BX,_mode
int $10
mov vysledek,AX
end;
if (vysledek<>$004F) then chyba('Nepovedlo se spustit graficky rezim.',0);
_setdefaultwindow;
CurBank:=0;{po inicializaci je nastavena banka 0}
End;{_setmode}

(***************************************************************************
                           PutPixel a GetPixel
****************************************************************************)

procedure _PutPixel(x,y:word;color:byte);assembler;
Asm
mov AX,__lhy
add y,AX       {pro virtualni obrazovku}
mov AX,$A000
mov ES,AX      {ES = segment pro videopamet}
mov AX,_MaxX
inc AX         {AX = sirka obrazovky}
mul y          {vynasob AX ypsilonem, co pretece, dej do DX}
add AX,x       {pricti x}
adc DX,0       {jestli AX preteklo, zvys hodnotu DX}
               {vysledek: v AX je ofset bodu, v DX cislo banky}
mov DI,AX      {ofset ulozime do registru k tomu urceneho}
               {ted se postarame o bankovani:}
cmp DX,CurBank   {neni nahodou potrebna banka uz nastavena?}
je @neprepinej    {je -> preskocime prepinani}
 mov curbank,DX      {nastavime promennou Curbank pro priste}
 mov AX,DX
 mul Granularity     {vypocet skutecneho cislo banky (nutne pri granularite <> 64 kB)}
 mov DX,AX           {__switchbank potrebuje cislo banky v DX, takze ho tam ulozime}
 mov AX,$4F05        {kod sluzby - "bankovani"}
 xor BX,BX           {BX:=0 ("nastav banku")}
 call __switchbank   {zavolame proceduru, ktera to provede}
@neprepinej:
mov AL,color
mov ES:[DI],AL     {samotne vykresleni pixelu uz je vylozene prkotina :-)}
End;{_putpixel}

procedure _PutTPixel(x,y:word;color:byte);assembler;
Asm
mov AL,color
or AL,AL
jz @Exit        {kdyz je color=0, koncime, jinak to same jako u _putpixelu}
 mov AX,__lhy
 add y,AX
 mov AX,$A000
 mov ES,AX
 mov AX,_MaxX
 inc AX
 mul y
 add AX,x
 adc DX,0
 mov DI,AX
 cmp DX,CurBank
 je @@1
  mov curbank,DX
  mov AX,DX
  mul Granularity
  mov DX,AX
  mov AX,$4F05
  xor BX,BX
  call __switchbank
 @@1:
 mov AL,color
 mov ES:[DI],AL
@Exit:
End;{_puttpixel}

procedure _XorPixel(x,y:word;color:byte);assembler;
Asm
mov AX,__lhy
add y,AX
mov AX,$A000
mov ES,AX
mov AX,_MaxX
inc AX
mul y
add AX,x
adc DX,0
mov DI,AX
cmp DX,CurBank
je @neprepinej
 mov curbank,DX
 mov AX,DX
 mul Granularity
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank
@neprepinej:
mov AL,color
xor byte ptr ES:[DI],AL
End;{_xorpixel}

function _GetPixel(x,y:word):byte; assembler;
Asm
mov AX,__lhy
add y,AX
mov AX,$A000
mov ES,AX
mov AX,_MaxX
inc AX
mul y
add AX,x
adc DX,0
mov DI,AX
cmp DX,CurBank
je @neprepinej
 mov curbank,DX
 mov AX,DX
 mul Granularity
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank
@neprepinej:
mov AL,ES:[DI]   {jediny rozdil proti _putpixelu je tady}
End;{_getpixel}

(***************************************************************************
                     PutPixel a GetPixel s orezavanim
****************************************************************************)

procedure __PutPixel(x,y:integer;color:byte);
Begin
if (x>=_MinClipX) and (x<=_MaxClipX) and
   (y>=_MinClipY) and (y<=_MaxClipY) then _PutPixel(x,y,color);
End;{__PutPixel}

procedure __PutTPixel(x,y:integer;color:byte);
Begin
if (color<>0) and
   (x>=_MinClipX) and (x<=_MaxClipX) and
   (y>=_MinClipY) and (y<=_MaxClipY) then _PutPixel(x,y,Color);
End;{__PutTPixel}

procedure __XorPixel(x,y:integer;color:byte);
Begin
if (x>=_MinClipX) and (x<=_MaxClipX) and
   (y>=_MinClipY) and (y<=_MaxClipY) then _xorpixel(x,y,color);
End;{__xorpixel}

function __GetPixel(x,y:integer):byte;
Begin
if x<_minclipx then x:=_minclipx
               else if x>_maxclipx then x:=_maxclipx;
if y<_minclipy then y:=_minclipy
               else if y>_maxclipy then y:=_maxclipy;
__getpixel:=_GetPixel(x,y);
End;{__getpixel}

(***************************************************************************
                           Kreslici procedury
****************************************************************************)

procedure _HLine(x1,x2,y:word;color:word); assembler;
var count:word;
Asm
{$ifdef blbuvzdornost}  {jestli x1>x2, prohodime je}
mov AX,x1
cmp AX,x2
js @OK
 mov AX,x1
 mov BX,x2
 mov x1,BX
 mov x2,AX
@OK:
{$endif}
mov AX,__lhy
add y,AX         {pro virtualni obrazovku}

mov AX,color
mov AH,AL
mov color,AX     {color ma ted hodnotu barvy v dolnim i hornim bytu}

mov AX,$A000
mov ES,AX        {ES = cilovy segment}

mov AX,x2
sub AX,x1
inc AX
mov count,AX     {count = pocet pixelu (delka cary)}

mov AX,_MaxX
inc AX
mul y
add AX,x1
adc DX,0         {DX = pocatecni banka}
mov DI,AX        {DI = ofset zacatku cary}

cmp DX,CurBank
je @neprepinej
 mov curbank,DX
 mov AX,DX
 mul Granularity
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank
 mov AX,DI
@neprepinej:        {pocatecni banka nastavena}

mov BX,count
dec BX
add AX,BX        {AX = ofset konce cary}
jnc @najednou     {kdyz se vejdeme do jedne banky, jdeme dal...}

 {...jinak vyplnime nejdriv konec prvni banky:}

 or CX,$FFFF         {= mov CX,$FFFF}
 sub CX,DI
 inc CX            {CX = pocet pixelu v prvni bance}

 mov AX,count
 sub AX,CX
 mov count,AX      {count = pocet pixelu ve druhe bance}

 mov AX,color
 db $66; shl AX,16
 mov AX,color      {EAX ma ted v kazdem bytu hodnotu barvy}

 mov BX,CX
 shr CX,2
 db $66; rep stosw  {= rep stosd, kopiruje se po 4 bytech}
 and BX,3
 mov CX,BX
 rep stosb          {prekopirovani zbylych max. 3 bytu}

 {v posledni iteraci DI preteklo a je v nem 0}

 inc curbank
 mov AX,curbank
 mul granularity   {v DX necha __switchbank OBCAS nejakou kravinu, tak musime znova nasobit a nejde jenom pricist granularitu}
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank   {jsme o banku dal}

@najednou:

mov CX,count         {CX = pocet pixelu (ve druhe bance nebo cele cary)}

mov AX,color
db $66; shl AX,16
mov AX,color      {EAX = barva}

mov BX,CX
shr CX,2
db $66; rep stosw
and BX,3
mov CX,BX
rep stosb              {zbytek cary je vykreslen}
End;{_hline}

procedure _VLine(x,y1,y2:word;color:byte);assembler;
var incr:word;
Asm
{$ifdef blbuvzdornost}  {jestli y1>y2, prohodime je}
mov AX,y1
cmp AX,y2
js @OK
 mov AX,y1
 mov BX,y2
 mov y1,BX
 mov y2,AX
@OK:
{$endif}

mov AX,__lhy         {pro virtualni obrazovku}
add y1,AX
add y2,AX

mov AX,$A000
mov ES,AX           {ES = $A000}

mov AX,_MaxX
inc AX
mov incr,AX         {AX = incr = sirka obrazovky}

mul y1
add AX,x
adc DX,0            {DX = pocatecni banka}
mov DI,AX           {DI = pocatecni ofset}

cmp DX,CurBank
je @neprepinej
 mov AX,DX
 mov curbank,DX
 mul granularity
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank
@neprepinej:          {pocatecni banka nastavena}

mov AL,color
mov ES:[DI],AL        {prvni pixel je nakreslen}

mov CX,y2
sub CX,y1             {CX = pocet cyklu}
jz @konec

 @cyklus:
 add DI,incr          {DI = ofset dalsiho pixelu}

 jnc @PoradVJedneBance
  inc curbank
  mov AX,curbank
  mul granularity
  mov DX,AX
  mov AX,$4F05
  xor BX,BX
  call __switchbank
  mov AL,color        {do AL vratime barvu}
 @PoradVJedneBance:   {spravna banka je nastavena}

 mov ES:[DI],AL       {pixel je nakreslen}
 loop @cyklus

@konec:
End;{_vline}

procedure _XorVLine(x,y1,y2:word;color:byte); assembler;
var incr:word;
Asm
{$ifdef blbuvzdornost}  {jestli y1>y2, prohodime je}
mov AX,y1
cmp AX,y2
js @OK
 mov AX,y1
 mov BX,y2
 mov y1,BX
 mov y2,AX
@OK:
{$endif}

mov AX,__lhy         {pro virtualni obrazovku}
add y1,AX
add y2,AX

mov AX,$A000
mov ES,AX           {ES = $A000}

mov AX,_MaxX
inc AX
mov incr,AX         {AX = incr = sirka obrazovky}

mul y1
add AX,x
adc DX,0            {DX = pocatecni banka}
mov DI,AX           {DI = pocatecni ofset}

cmp DX,CurBank
je @neprepinej
 mov AX,DX
 mov curbank,DX
 mul granularity
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank
@neprepinej:          {pocatecni banka nastavena}

mov AL,color
xor byte ptr ES:[DI],AL        {prvni pixel je nakreslen}

mov CX,y2
sub CX,y1             {CX = pocet cyklu}
jz @konec

 @cyklus:
 add DI,incr          {DI = ofset dalsiho pixelu}

 jnc @PoradVJedneBance
  inc curbank
  mov AX,curbank
  mul granularity
  mov DX,AX
  mov AX,$4F05
  xor BX,BX
  call __switchbank
  mov AL,color        {do AL vratime barvu}
 @PoradVJedneBance:   {spravna banka je nastavena}

 xor byte ptr ES:[DI],AL       {pixel je nakreslen}
 loop @cyklus

@konec:
End;{_xorvline}

procedure _XorHLine(x1,x2,y:word;color:word); assembler;
var count:word;
Asm
{$ifdef blbuvzdornost}  {jestli x1>x2, prohodime je}
mov AX,x1
cmp AX,x2
js @OK
 mov AX,x1
 mov BX,x2
 mov x1,BX
 mov x2,AX
@OK:
{$endif}
mov AX,__lhy
add y,AX         {pro virtualni obrazovku}

mov AX,color
mov AH,AL
mov color,AX     {color ma ted hodnotu barvy v dolnim i hornim bytu}

mov AX,$A000
mov ES,AX        {ES = cilovy segment}

mov AX,x2
sub AX,x1
inc AX
mov count,AX     {count = pocet pixelu (delka cary)}

mov AX,_MaxX
inc AX
mul y
add AX,x1
adc DX,0         {DX = pocatecni banka}
mov DI,AX        {DI = ofset zacatku cary}

cmp DX,CurBank
je @neprepinej
 mov curbank,DX
 mov AX,DX
 mul Granularity
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank
 mov AX,DI
@neprepinej:        {pocatecni banka nastavena}

mov BX,count
dec BX
add AX,BX        {AX = ofset konce cary}
jnc @najednou     {kdyz se vejdeme do jedne banky, jdeme dal...}

 {...jinak vyplnime nejdriv konec prvni banky:}

 or CX,$FFFF         {= mov CX,$FFFF}
 sub CX,DI
 inc CX            {CX = pocet pixelu v prvni bance}

 mov AX,count
 sub AX,CX
 mov count,AX      {count = pocet pixelu ve druhe bance}

 mov AX,color
 db $66; shl AX,16
 mov AX,color      {EAX ma ted v kazdem bytu hodnotu barvy}

 mov BX,CX
 shr CX,2

 jz @nic1  {aby se nekreslilo nic, pokud CX=0 (o to se normalne stara Rep)}
  @smycka1:
  db $66; xor ES:[DI],AX
(*  db $66; mov DX,ES:[DI]
  db $66; xor DX,AX   {xorovani po dwordech}
  db $66; mov dword ptr ES:[DI],AX*)
  add DI,4
  loop @smycka1
 @nic1:

 and BX,3
 mov CX,BX

 or CX,CX
 jz @nic2
  @smycka2:
  xor byte ptr ES:[DI],AL  {prexorovani zbylych max. 3 bytu}
  inc DI
  loop @smycka2
 @nic2:
 {v posledni iteraci DI preteklo a je v nem 0}

 inc curbank
 mov AX,curbank
 mul granularity   {v DX nechava __switchbank OBCAS kraviny, tak musime znova nasobit a nejde jenom pricist granularitu}
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank   {jsme o banku dal}

@najednou:

mov CX,count         {CX = pocet pixelu (ve druhe bance nebo cele cary)}

mov AX,color
db $66; shl AX,16
mov AX,color      {EAX = barva}

mov BX,CX
shr CX,2
jz @nic3
 @smycka3:
 db $66; xor ES:[DI],AX   {dword ptr}
 add DI,4
 loop @smycka3
@nic3:
and BX,3
mov CX,BX
or CX,CX
jz @nic4
 @smycka4:
 xor byte ptr ES:[DI],AL
 inc DI
 loop @smycka4
@nic4:
End;{_xorhline}

procedure _line(x1,y1,x2,y2:word; barva:byte);
var dx,dy,krokx,kroky:integer;
    pom:word;
Begin
{Bresenhamuv algoritmus obecne cary:
K pomocne promenne pricitame hodnotu mensiho z rozdilu souradnic (dx nebo dy)
a ridici promennou pri tom menime o 1. Jakmile pomocna promenna dosahne nebo
presahne hodnotu vetsiho z rozdilu souradnic, ten vetsi rozdil od ni odecteme
a zmenime o 1 i druhou souradnici.
Na zacatku musi mit pomocna promenna hodnotu rovnou polovine toho vetsiho
rozdilu souradnic, jinak by cara byla nesymetricka.}
{Pokud vychazi cara vodorovna nebo svisla, pouzijeme z duvodu rychlosti radsi
procedury pro vodorovnou nebo svislou caru, i kdyz Bresenhamuv algoritmus by
fungoval i v techto pripadech (krok druhe souradnice by byl 0):}
if x1=x2 then begin
              if y2>y1 then _vline(x1,y1,y2,barva)
                       else _vline(x1,y2,y1,barva);
              Exit;
              end;
if y1=y2 then begin
              if x2>x1 then _hline(x1,x2,y1,barva)
                       else _hline(x2,x1,y1,barva);
              Exit;
              end;
{vypocet rozdilu souradnic a kroku:}
if x2>x1 then begin dx:=x2-x1; krokx:=1; end
         else begin dx:=x1-x2; krokx:=-1; end;
if y2>y1 then begin dy:=y2-y1; kroky:=1; end
         else begin dy:=y1-y2; kroky:=-1; end;
{Zadny rozdil souradnic urcite neni 0, protoze to by odchytily predchozi dve
podminky a procedura by uz touhle dobou davno skoncila. Takze se ted o takovy
pripad nemusime starat.}
_putpixel(x1,y1,barva); {prvni bod cary}
if dx>=dy then begin {cara je spis vodorovnejsi, x je ridici promenna}
               pom:=dx shr 1; {pocatecni hodnota pomocne promenne}
                repeat
                inc(x1,krokx); {zmenime x o 1}
                inc(pom,dy); {pricteme k pomocne dy (mensi rozdil)}
                if pom>=dx then begin {kdyz je vesti nez dx (ten vetsi rozdil)...}
                                dec(pom,dx); {...vetsi rozdil odecteme...}
                                inc(y1,kroky); {...a zmenime y o 1}
                                end;
                _putpixel(x1,y1,barva); {nakreslime bod}
                until x1=x2; {dokud nejsme na koncovem bodu}
               end
          else begin {cara je spis svislejsi, y je ridici}
               pom:=dy shr 1;
                repeat         {postup stejny, jen jsou prehozene souradnice}
                inc(y1,kroky);
                inc(pom,dx);
                if pom>=dy then begin
                                dec(pom,dy);
                                inc(x1,krokx);
                                end;
                _putpixel(x1,y1,barva);
                until y1=y2;
               end;
End;{_line}

procedure _XorLine(x1,y1,x2,y2:word;barva:byte);
var dx,dy,krokx,kroky:integer;
    pom:word;
Begin
if x1=x2 then begin
              if y2>y1 then _xorvline(x1,y1,y2,barva)
                       else _xorvline(x1,y2,y1,barva);
              Exit;
              end;
if y1=y2 then begin
              if x2>x1 then _xorhline(x1,x2,y1,barva)
                       else _xorhline(x2,x1,y1,barva);
              Exit;
              end;
if x2>x1 then begin dx:=x2-x1; krokx:=1; end
         else begin dx:=x1-x2; krokx:=-1; end;
if y2>y1 then begin dy:=y2-y1; kroky:=1; end
         else begin dy:=y1-y2; kroky:=-1; end;
_xorpixel(x1,y1,barva);
if dx>=dy then begin
               pom:=dx shr 1;
                repeat
                inc(x1,krokx);
                inc(pom,dy);
                if pom>=dx then begin
                                dec(pom,dx);
                                inc(y1,kroky);
                                end;
                _xorpixel(x1,y1,barva);
                until x1=x2;
               end
          else begin
               pom:=dy shr 1;
                repeat
                inc(y1,kroky);
                inc(pom,dx);
                if pom>=dy then begin
                                dec(pom,dy);
                                inc(x1,krokx);
                                end;
                _xorpixel(x1,y1,barva);
                until y1=y2;
               end;
End;{_xorline}

procedure _MaskedLine(x1,y1,x2,y2:integer; barva:byte; maska:word);
var dx,dy,krokX,krokY:integer;
    pom:word;
Begin
{Tady uz se neda pouzit rychla vodorovna cara, takze pro vsechny pripady
pouzijeme obecny Bresenhamuv algoritmus.}
dx:=abs(x2-x1); dy:=abs(y2-y1);
if x2<x1 then krokx:=-1
         else if x2>x1 then krokx:=1
                       else krokx:=0;
if y2<y1 then kroky:=-1
         else if y2>y1 then kroky:=1
                       else kroky:=0;
{prvni bod cary:}
asm
rol maska,1 {rotujeme masku o jeden bit doleva}
jnc @dira1  {kdyz nenastalo preteceni, odrotovali jsme nulu, takze se nic nekresli}
 push x1              {ulozime prvni parametr pro _putpixel - x}
 push y1              {druhy parametr - y}
 push word ptr barva  {treti parametr - barva (na zasobnik jdou ukladat jen wordy, tak se musi pretypovat)}
 call _putpixel       {zavolane _putpixel}
@dira1:
end;
if dx>=dy then begin
               pom:=dx shr 1;
                repeat
                inc(x1,krokx);
                inc(pom,dy);
                if pom>=dx then begin
                                dec(pom,dx);
                                inc(y1,kroky);
                                end;
                asm
                rol maska,1
                jnc @dira2
                 push x1
                 push y1
                 push word ptr barva
                 call _putpixel
                @dira2:
                end;
                until x1=x2;
               end
          else begin
               pom:=dy shr 1;
                repeat
                inc(y1,kroky);
                inc(pom,dx);
                if pom>=dy then begin
                                dec(pom,dy);
                                inc(x1,krokx);
                                end;
                asm
                rol maska,1
                jnc @dira3
                 push x1
                 push y1
                 push word ptr barva
                 call _putpixel
                @dira3:
                end;
                until y1=y2;
               end;
End;{_maskedline}

procedure _xorMaskedLine(x1,y1,x2,y2:integer; barva:byte; maska:word);
var dx,dy,krokX,krokY:integer;
    pom:word;
Begin
dx:=abs(x2-x1);
dy:=abs(y2-y1);
if x2<x1 then krokx:=-1
         else if x2>x1 then krokx:=1
                       else krokx:=0;
if y2<y1 then kroky:=-1
         else if y2>y1 then kroky:=1
                       else kroky:=0;
asm
rol maska,1
jnc @dira1
 push x1
 push y1
 push word ptr barva
 call _xorpixel
@dira1:
end;
if dx>=dy then begin
               pom:=dx shr 1;
                repeat
                inc(x1,krokx);
                inc(pom,dy);
                if pom>=dx then begin
                                dec(pom,dx);
                                inc(y1,kroky);
                                end;
                asm
                rol maska,1
                jnc @dira2
                 push x1
                 push y1
                 push word ptr barva
                 call _xorpixel
                @dira2:
                end;
                until x1=x2;
               end
          else begin
               pom:=dy shr 1;
                repeat
                inc(y1,kroky);
                inc(pom,dx);
                if pom>=dy then begin
                                dec(pom,dy);
                                inc(x1,krokx);
                                end;
                asm
                rol maska,1
                jnc @dira3
                 push x1
                 push y1
                 push word ptr barva
                 call _xorpixel
                @dira3:
                end;
                until y1=y2;
               end;
End;{_xormaskedline}

procedure _Bar(x1,y1,x2,y2:word;barva:word);assembler;
var sirka,incr:word;
Asm
{$ifdef blbuvzdornost}
mov AX,x1
cmp AX,x2
js @xOK
 mov AX,x1
 mov BX,x2              {jestli x1>x2, prohodime je}
 mov x1,BX
 mov x2,AX
@xOK:
mov AX,y1
cmp AX,y2
js @yOK
 mov AX,y1
 mov BX,y2              {jestli y1>y2, prohodime je}
 mov y1,BX
 mov y2,AX
@yOK:
{$endif}

mov AX,__lhy
add y1,AX
add y2,AX         {pro virtualni obrazovku}

mov AX,$A000
mov ES,AX        {ES = cilovy segment}

inc y2         {aby se to pak lip pocitalo}

mov AX,barva
mov AH,AL
mov barva,AX     {barva ma ted hodnotu v obou bytech}

mov AX,_maxx
sub AX,x2
add AX,x1
mov incr,AX      {incr = kolik se musi pridat k ofsetu na konci radku, abychom se dostali na zacatek dalsiho}

mov AX,x2
sub AX,x1
inc AX
mov sirka,AX         {sirka spocitana}

mov AX,_Maxx
inc AX
mul y1
add AX,x1
adc DX,0        {DX = banka}
mov DI,AX       {DI = ofset zacatku radku}

 @zacatek:         {zacatek cyklu}

 mov BX,DI
 mov CX,DX
 add BX,sirka
 adc CX,0
 dec BX
 sbb CX,0           {CX = banka bodu na konci radku}

 cmp DX,curbank
 je @neprepinej
  push DX
  call _SwitchBank
 @neprepinej:         {pocatecni banka nastavena}

 cmp CX,curbank       {je banka na konci radku stejna jako na zacatku?}
 je @najednou         {je -> vykresli se cely radek najednou}

  or CX,$FFFF
  sub CX,DI
  inc CX            {CX = sirka prvni casti}
  mov AX,sirka
  sub AX,CX
  push AX     {zasobnik = sirka druhe casti}

  mov AX,barva
  mov BX,AX
  db $66; shl AX,16
  mov AX,BX           {EAX = barva}

   mov BX,CX
   shr CX,2
   db $66; rep stosw
   and BX,3
   mov CX,BX
   rep stosb

  mov AX,curbank
  inc AX
  push AX
  call _switchbank      {druha banka nastavena}

  pop CX
  jmp @jedeme

 @najednou:
 mov CX,sirka
 @jedeme:
                       {CX = sirka zbytku radku}
 mov AX,barva
 mov BX,AX
 db $66; shl AX,16
 mov AX,BX           {EAX = barva}

  mov BX,CX
  shr CX,2
  db $66; rep stosw
  and BX,3
  mov CX,BX
  rep stosb

 mov DX,curbank
 add DI,incr       {DI = ofset zacatku dalsiho radku}
 adc DX,0          {DX = banka -''-}
 inc y1
 mov AX,y1
 cmp AX,y2
 jne @zacatek
End;{_bar}

procedure _MaskedBar(x1,y1,x2,y2,barva:word;var vzorek:_pole8bytu);
var i,sirka,sirka2:word;
    m:byte;
    pom:array[0..3] of word; {cili qword}
    poms,pomo:word;{segment a ofset pole pom}
Begin
{$ifdef blbuvzdornost}
asm
mov AX,x1
cmp AX,x2
js @xOK
 mov AX,x1
 mov BX,x2              {jestli x1>x2, prohodime je}
 mov x1,BX
 mov x2,AX
@xOK:
mov AX,y1
cmp AX,y2
js @yOK
 mov AX,y1
 mov BX,y2              {jestli y1>y2, prohodime je}
 mov y1,BX
 mov y2,AX
@yOK:
end;
{$endif}
poms:=seg(pom); pomo:=ofs(pom);{v asm mi to nejak neslo}
sirka:=x2-x1+1;
inc(y1,__lhy); inc(y2,__lhy);
for i:=y1 to y2 do
  begin
  m:=vzorek[i and 7];
  asm{vykresleni radku}
  mov CX,x1
  and CX,7
  rol m,CL      {upraveni masky podle souradnic - aby navazovala}

  mov AX,poms
  mov ES,AX
  mov DI,pomo   {ES:DI = adresa pom}

  mov AX,barva  {AL = barva popredi, AH = barva pozadi}
  mov CX,8
   @PripravaDat:
   rol m,1
   jc @BarvaPopredi
    mov ES:[DI],AH         {pixel v barve pozadi}
    jmp @KonecVyberuBarvy
   @BarvaPopredi:
    mov ES:[DI],AL         {pixel v barve popredi}
   @KonecVyberuBarvy:
   inc DI
   loop @PripravaDat
                          {ted je v Pom jeden radek masky, co byte, to pixel}
  mov AX,$A000
  mov ES,AX       {ES = cilovy segment}

  mov AX,_Maxx
  inc AX
  mul i
  add AX,x1
  adc DX,0     {DX = banka zacatku radku}
  mov DI,AX    {DI = ofset zacatku radku}

  mov BX,DI
  mov CX,DX
  add BX,sirka
  adc CX,0
  dec BX         {BX = ofset bodu na konci radku}
  sbb CX,0       {CX = banka toho bodu}

  {ted je v DI ofset zacatku radku, v DX jeho banka a v BX a CX to same pro konec radku
  (ten ofset v BX nepotrebuju a stejne ho Switchbank prepise)}

  cmp DX,curbank  {prepnuti do banky pro zacatek radku}
  je @neprepinej
   push DX
   call _SwitchBank
  @neprepinej:

  cmp CX,curbank {je banka na konci radku stejna jako na zacatku?}
  je @najednou  {je => hop dal ***}
   {neni => nakresli se prvni cast radku v prvni bance, pak se prepne o banku dal a nakresli se zbytek:}

   or CX,$FFFF  {= mov CX,$FFFF, ale trochu rychlejsi}
   sub CX,DI
   inc CX        {CX = pocet pixelu v prvni bance}

   mov AX,sirka
   sub AX,CX
   mov sirka2,AX   {sirka2 = pocet pixelu ve druhe bance}

   {kresleni po qwordech (8 bytu najednou - Need for Speed hadr :-) ):}
   mov BX,CX
   shr CX,3
   jz @nic1
    @PoQwordech1:
    fild qword ptr pom           {kopiruj pole na zasobnik koprocesoru}
    fistp qword ptr ES:[DI]      {presun pole ze zasobniku na obrazovku}
    add DI,8
    loop @PoQwordech1
   @nic1:
   {zbytek po bytech:}
   mov DX,DS
   mov AX,barva
   mov CX,BX
   and CX,7
   push CX  {schovat na potom}
   mov AX,poms
   mov DS,AX
   mov SI,pomo
    rep movsb
   mov DS,DX
   {v posledni iteraci DI preteklo a je v nem 0, coz presne chceme}

   inc curbank
   push curbank
   call _switchbank     {prepnuti o banku dal (DI je 0, takze budeme kreslit na zacatek dalsi banky)}

   {odrotovani masky, aby druha cast radku navazovala na prvni:}
   pop CX
   or CL,CL
   jz @NetrebaRotovati
    rol m,CL
    mov AX,poms
    mov ES,AX
    mov BX,pomo   {ES:BX = adresa pom}
    mov AX,barva
    mov CX,8
     @PripravaDat2:
     rol m,1
     jc @BarvaPopredi2
      mov ES:[BX],AH         {pixel v barve pozadi}
      jmp @KonecVyberuBarvy2
     @BarvaPopredi2:
      mov ES:[BX],AL         {pixel v barve popredi}
     @KonecVyberuBarvy2:
     inc BX
     loop @PripravaDat2
    mov AX,$A000
    mov ES,AX
   @NetrebaRotovati:

   mov CX,sirka2  {CX = pocet pixelu ve druhe bance}
   jmp @jedeme

  @najednou:
  mov CX,sirka  {CX = pocet pixelu}

  @jedeme:

  {kresleni:}
  mov BX,CX
  shr CX,3
  jz @nic2
   @PoQwordech2:
   fild qword ptr pom
   fistp qword ptr ES:[DI]
   add DI,8
   loop @PoQwordech2
  @nic2:
  mov DX,DS
  mov AX,barva
  mov CX,BX
  and CX,7
  mov AX,poms
  mov DS,AX
  mov SI,pomo
   rep movsb
  mov DS,DX
  {radek je vykreslen}
  end;
  end;
End;{_maskedbar}

procedure _Rectangle(x1,y1,x2,y2:word;color:byte);
Begin
_hline(x1,x2,y1,color);
_vline(x1,y1,y2,color);
_vline(x2,y1,y2,color);
_hline(x1,x2,y2,color);
End;{_rectangle}

procedure _XorRectangle(x1,y1,x2,y2:word;color:byte);
Begin
_xorhline(x1,x2,y1,color);
_xorvline(x1,y1,y2,color);
_xorvline(x2,y1,y2,color);
_xorhline(x1,x2,y2,color);
End;{_xorrectangle}

procedure _Box(x1,y1,x2,y2:word;lhbarva,pdbarva,vbarva:word);
Begin
_hline(x1,x2,y1,lhbarva);
_vline(x1,y1,y2,lhbarva);
_vline(x2,y1,y2,pdbarva);
_hline(x1,x2,y2,pdbarva);
if (vbarva<>0)and(x2-x1>=2)and(y2-y1>=2) then _bar(x1+1,y1+1,x2-1,y2-1,vbarva);
End;{_box}

{puvodni asm podoba zrusena (docasne?) kvuli problemum pri strankovani:}
(*assembler;
var bank,ofset,count,incr:word;
Asm
{$ifdef blbuvzdornost}
mov AX,x1
cmp AX,x2
js @xOK
 mov AX,x1
 mov BX,x2
 mov x1,BX
 mov x2,AX
@xOK:
mov AX,y1
cmp AX,y2
js @yOK
 mov AX,y1
 mov BX,y2
 mov y1,BX
 mov y2,AX
@yOK:
{$endif}
 mov ax,_MaxX
 inc ax
 mov incr,ax
 cld

 push x1                        { Draw top line of box }
 push x2
 push y1
 push word ptr lhbarva
 call _HLine

 push x1                        { Draw left and right sides }
 push y1
 push y2
 push word ptr lhbarva
 call _VLine

 push x2
 push y1
 push y2
 push word ptr pdbarva
 call _VLine

 mov AX,__lhy
 add y1,AX
 add y2,AX        {pro virtualni obrazovku}

 mov ax,$A000   { Set up ES:0 for videobuffer segment }
 mov es,ax
 mov ax,x2      { Check if zero-width box }
 sub ax,x1
 jz @@Exit
 dec ax
 jz @@Exit
 mov count,ax
 inc x1
 mov ax,y2      { Check if zero-height box }
 sub ax,y1
 jz @@Exit
 dec ax
 jz @@4
 inc y1

 mov AX,vbarva
 or AX,AX                       { filled rectangle? }
 jz @@4                         { if no, go to bottom line drawing }
 mov AH,AL
 mov vbarva,AX         {zkopirovani barvy na cely word}

 mov ax,_MaxX
 inc ax
 mul y1
 add ax,x1
 adc dx,0
 mov bank,dx
 mov ofset,ax
 cmp dx,CurBank
 je @@entry
 push bank
 mov cx,ax
 call SwitchBank
 mov ax,cx
 jmp @@entry

@@7:
 mov ax,ofset
 add ax,incr
 mov ofset,ax
 jnc @@3
 inc bank
 push bank
 call SwitchBank
@@3:

@@entry:
 add ax,count
 jc @@5                { if next bank, goto @@5 }
 mov di,ofset          { if no, fill line }
 mov cx,count
 mov AX,vbarva
 db $66; shl ax,16     { shl eax,16 }
 mov ax,vbarva
 mov bx,cx
 shr cx,2
 db 66h; rep stosw     { rep stosd }
 and bx,3
 mov cx,bx
 rep stosb
 jmp @@Exit2
@@5:
 mov di,ofset          { fill part in current bank }
 or cx,$FFFF           { mov cx,$FFFF }
 sub cx,di
 inc cx
 mov AX,vbarva
 db $66; shl ax,16
 mov AX,vbarva
 mov bx,cx
 shr cx,2
 db 66h; rep stosw
 and bx,3
 mov cx,bx
 rep stosb
 inc bank
 push bank             { switch to next bank }
 call SwitchBank
 dec bank
 xor di,di             { and fill the rest of horizontal line }
 mov cx,ofset
 add cx,count
 mov AX,vbarva
 db $66; shl ax,16
 mov AX,vbarva
 mov bx,cx
 shr cx,2
 db 66h; rep stosw
 and bx,3
 mov cx,bx
 rep stosb
@@Exit2:
 inc y1
 mov ax,y2
 cmp ax,y1              { while y<y2 }
 jne @@7

@@4:                    { Draw bottom line }
 dec x1
 push x1
 push x2
 push y2
 push word ptr pdbarva
 call _Hline
@@Exit:                 { Voila! }
End;{_box}*)

procedure _MaskedBox(x1,y1,x2,y2:word;lhbarva,pdbarva,vbarva:word;var vzorek:_pole8bytu);
Begin
{$ifdef blbuvzdornost}
asm
mov AX,x1
cmp AX,x2
js @xOK
 mov AX,x1
 mov BX,x2
 mov x1,BX
 mov x2,AX
@xOK:
mov AX,y1
cmp AX,y2
js @yOK
 mov AX,y1
 mov BX,y2
 mov y1,BX
 mov y2,AX
@yOK:
end;
{$endif}
_hline(x1,x2,y1,lhbarva);
_vline(x1,y1,y2,lhbarva);
_vline(x2,y1,y2,pdbarva);
_hline(x1,x2,y2,pdbarva);
if (vbarva<>0)and(x2-x1>=2)and(y2-y1>=2) then _maskedbar(x1+1,y1+1,x2-1,y2-1,vbarva,vzorek);
End;{_MaskedBox}

procedure _Fill(color:word);assembler;
var nbanks,bank,rest:word;
Asm
mov AX,color
mov AH,AL
mov color,AX   {barva je rozkopirovana na cely word}
mov BX,_MaxX
inc BX
mov AX,_MaxY
inc AX
mul BX           {sirka * vyska obrazovky}
mov nbanks,DX    {pretekla cast = pocet bank}
mov rest,AX      {co nepreteklo, to je pocet pixelu v posledni bance}
mov AX,$A000
mov ES,AX        {ES = segment obrazovky}
xor AX,AX
mov bank,AX      {zacneme na bance 0}
@@1:
 push bank
 call _SwitchBank    {banka nastavena}
 xor DI,DI           {DI na zacatek banky}
 mov CX,$4000        {banka ma 64 kB ($10000 B), tohle je $10000/4}
 mov AX,color
 mov BX,AX
 db $66; shl AX,16   {shl EAX,16}
 mov AX,BX           {EAX = barva v kazdem bytu}
 db $66; rep stosw   {rep stosd (kopirovani po 4 bytech)}
 inc bank            {dalsi banka}
 mov AX,bank
 dec AX
 cmp AX,nbanks
 jne @@1       {kdyz jeste nejsme u te posledni, ktera neni na obrazovce cela, hop zpet}
push bank
call _SwitchBank
xor DI,DI
mov CX,rest     {pocet pixelu v posledni bance}
inc CX
mov AX,color
mov BX,AX
db $66; shl AX,16     {EAX = barva}
mov AX,BX
mov BX,CX
shr CX,2
db $66; rep stosw     {rep stosd dokud to jde}
and BX,3
mov CX,BX
rep stosb             {rep stosb pro posledni max. 3 byty}
End;{_fill}

procedure _Fill2(color:word);
Begin
_bar(0,0,_maxx,_maxy,color);
End;{_fill2}

procedure _Ellipse(xc,yc,rx,ry:word;color:byte);
var x,y:word;
    a,b,as,tas,bs,tbs:longint;
    d,dx,dy:longint;
Begin
if rx=0 then begin
             _VLine(xc,yc-ry,yc+ry,color);
             Exit;
             end;
if ry=0 then begin
             _HLine(xc-rx,xc+rx,yc,color);
             Exit;
             end;
x:=0; y:=ry; a:=rx; b:=ry; as:=a*a;
tas:=as shl 1; bs:=b*b; tbs:=bs shl 1;
d:=bs-as*b+(as shr 2); dx:=0; dy:=tas*b;
while dx<dy do begin
               _PutPixel(xc+x,yc+y,color);
               _PutPixel(xc-x,yc+y,color);
               _PutPixel(xc+x,yc-y,color);
               _PutPixel(xc-x,yc-y,color);
               if d>0 then begin y:=y-1; dy:=dy-tas; d:=d-dy; end;
               Inc(x); dx:=dx+tbs; d:=d+bs+dx;
               end;
d:=d+((3*(as-bs)div 2-(dx+dy))div 2);
while y>0 do begin
             _PutPixel(xc+x,yc+y,color);
             _PutPixel(xc-x,yc+y,color);
             _PutPixel(xc+x,yc-y,color);
             _PutPixel(xc-x,yc-y,color);
             if d<0 then begin Inc(x); dx:=dx+tbs; d:=d+dx; end;
             y:=y-1; dy:=dy-tas; d:=d+as-dy;
             end;
_PutPixel(xc+x,yc,color);
_PutPixel(xc-x,yc,color);
End;{_ellipse}

procedure _XorEllipse(xc,yc,rx,ry:word;color:byte);
var x,y:word;
    a,b,as,tas,bs,tbs:longint;
    d,dx,dy:longint;
Begin
if rx=0 then begin
             _xorVLine(xc,yc-ry,yc+ry,color);
             Exit;
             end;
if ry=0 then begin
             _xorHLine(xc-rx,xc+rx,yc,color);
             Exit;
             end;
x:=0; y:=ry; a:=rx; b:=ry; as:=a*a;
tas:=as shl 1; bs:=b*b; tbs:=bs shl 1;
d:=bs-as*b+(as shr 2); dx:=0; dy:=tas*b;
while dx<dy do begin
               _xorPixel(xc+x,yc+y,color);
               _xorPixel(xc-x,yc+y,color);
               _xorPixel(xc+x,yc-y,color);
               _xorPixel(xc-x,yc-y,color);
               if d>0 then begin y:=y-1; dy:=dy-tas; d:=d-dy; end;
               Inc(x); dx:=dx+tbs; d:=d+bs+dx;
               end;
d:=d+((3*(as-bs)div 2-(dx+dy))div 2);
while y>0 do begin
             _xorPixel(xc+x,yc+y,color);
             _xorPixel(xc-x,yc+y,color);
             _xorPixel(xc+x,yc-y,color);
             _xorPixel(xc-x,yc-y,color);
             if d<0 then begin Inc(x); dx:=dx+tbs; d:=d+dx; end;
             y:=y-1; dy:=dy-tas; d:=d+as-dy;
             end;
_xorPixel(xc+x,yc,color);
_xorPixel(xc-x,yc,color);
End;{_xorellipse}

procedure _FilledEllipse(xc,yc,rx,ry:word;color:byte);
var x,y:word;
    a,b,as,tas,bs,tbs:longint;
    d,dx,dy:longint;
Begin
if rx=0 then begin
             _VLine(xc,yc-ry,yc+ry,color);
             Exit;
             end;
if ry=0 then begin
             _HLine(xc-rx,xc+rx,yc,color);
             Exit;
             end;
x:=0; y:=ry; a:=rx; b:=ry; as:=a*a;
tas:=as shl 1; bs:=b*b; tbs:=bs shl 1;
d:=bs-as*b+(as shr 2); dx:=0; dy:=tas*b;
while dx<dy do begin
               _HLine(xc-x,xc+x,yc+y,color);
               _HLine(xc-x,xc+x,yc-y,color);
               if d>0 then begin y:=y-1; dy:=dy-tas; d:=d-dy; end;
               Inc(x); dx:=dx+tbs; d:=d+bs+dx;
               end;
d:=d+((3*(as-bs)div 2-(dx+dy))div 2);
while y>0 do begin
             _HLine(xc-x,xc+x,yc+y,color);
             _HLine(xc-x,xc+x,yc-y,color);
             if d<0 then begin Inc(x); dx:=dx+tbs; d:=d+dx; end;
             y:=y-1; dy:=dy-tas; d:=d+as-dy;
             end;
_HLine(xc-x,xc+x,yc,color);
End;{_filledellipse}

procedure _Triangle(x1,y1,x2,y2,x3,y3:word;color:byte);
Begin
_Line(x1,y1,x2,y2,color);
_Line(x2,y2,x3,y3,color);
_Line(x3,y3,x1,y1,color);
End;{_triangle}

procedure _xorTriangle(x1,y1,x2,y2,x3,y3:word;color:byte);
Begin
_xorLine(x1,y1,x2,y2,color);
_xorLine(x2,y2,x3,y3,color);
_xorLine(x3,y3,x1,y1,color);
End;{_xortriangle}

procedure xchange(var a,b:integer); assembler;
Asm
push DS
lds SI,a          {nacteni ukazatelu na promenne}
les DI,b
mov AX,DS:[SI]    {nacteni hodnot promennych}
mov BX,ES:[DI]
mov DS:[SI],BX    {ulozeni hodnot v obracenem poradi}
mov ES:[DI],AX
pop DS
End;{xchange}

procedure _FilledTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
var ymin,xmin,ymid,xmid,ymax,xmax,xmid2:longint;
    i:integer;
Begin
{srovnani vrcholu podle souradnice y:}
if y1>y2 then begin xchange(y1,y2); xchange(x1,x2); end;
if y1>y3 then begin xchange(y1,y3); xchange(x1,x3); end;
if y2>y3 then begin xchange(y2,y3); xchange(x2,x3); end;
ymin:=y1; xmin:=x1; ymid:=y2; xmid:=x2; ymax:=y3; xmax:=x3;
{kdyz jsou vsechny body stejne vysoko, je to vodorovna cara a ne trojuhelnik:}
if ymax=ymin then begin
                  {nalezeni xove souradnice nejvic vlevo a nejvic vpravo:}
                  if x1>x2 then xchange(x1,x2);
                  if x1>x3 then xchange(x1,x3);
                  if x2>x3 then xchange(x2,x3);
                  _hline(x1,x3,y1,color);
                  exit;
                  end;
{vrcholy jsou primo pod sebou - je to svisla cara:}
if (x1=x2)and(x2=x3) then begin
                          _vline(x1,y1,y3,color);
                          exit;
                          end;
xmid2:=(xmax-xmin)*(ymid-ymin)div(ymax-ymin)+xmin;
   {bod na care, ktera jde od nejhorejsiho vrcholu k nejdolejsimu, naproti prostrednimu vrcholu}
if xmid2<xmid then xchange(integer(xmid2),integer(xmid));
if ymin<>ymid then for i:=ymin to ymid do _HLine((xmid-xmin)*(i-ymin)div(ymid-ymin)+xmin,
                                                 (xmid2-xmin)*(i-ymin)div(ymid-ymin)+xmin,i,color);
if ymid<>ymax then for i:=ymid to ymax do _HLine((xmax-xmid)*(i-ymid)div(ymax-ymid)+xmid,
                                                 (xmax-xmid2)*(i-ymid)div(ymax-ymid)+xmid2,i,color);
End;{_filledtriangle}

(*procedure _FilledTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
var dxl,krokXl,krokYl,dxp,dy,krokXp,krokYp,xl,xp,yl,yp,y:integer;
    poml,pomp:word;
    stredX:integer; {bod na spojnici bodu 1 a 3, stredY = y2}
Begin
{srovnani vrcholu podle souradnice y:}
if y1>y2 then begin xchange(y1,y2); xchange(x1,x2); end;
if y1>y3 then begin xchange(y1,y3); xchange(x1,x3); end;
if y2>y3 then begin xchange(y2,y3); xchange(x2,x3); end;
{ted je 1 nejvys, 3 nejniz a 2 nekde mezi}
if y1=y3 then begin{je to vodorovna cara a ne trojuhelnik}
              {nalezeni xove souradnice nejvic vlevo a nejvic vpravo:}
              if x1>x2 then xchange(x1,x2);
              if x1>x3 then xchange(x1,x3);
              if x2>x3 then xchange(x2,x3);
              _hline(x1,x3,y1,color);
              exit;
              end;
if (x1=x2)and(x2=x3) then begin{vrcholy jsou pod sebou - je to svisla cara}
                          _vline(x1,y1,y3,color);
                          exit;
                          end;
{Kresleni trojuhelnika se deli na dve casti, horni od y1 do y2 a dolni od y2
do y3. Spojnici 1 - 3 rozdelime v bode [stredx,y2]:}
stredx:=longint(x3-x1)*(y2-y1)div(y3-y1)+x1;
{aby se to nepletlo, bude x2 vlevo a stredx vpravo:}
if stredx<x2 then xchange(stredx,x2);
if y1<>y2 then
 begin
 {horni cast trojuhelnika (dvojity Bresenhamuv algoritmus cary, vlevo je cara
 z x1,y1 do x2,y2 (index l), vpravo je cara z x1,y1 do stredx,y2 (index p)):}
 dxl:=abs(x2-x1);
 if x2<x1 then krokxl:=-1
          else if x2>x1 then krokxl:=1
                        else krokxl:=0;
 dxp:=abs(stredx-x1);
 if stredx<x1 then krokxp:=-1
              else if stredx>x1 then krokxp:=1
                                else krokxp:=0;
 dy:=abs(y2-y1);{pro obe cary spolecna}
 {krokY je diky serazeni vrcholu pro obe cary roven 1}
 xl:=x1; xp:=x1; yl:=y1; yp:=y1; y:=y1;{pomocne promenne, ktere se budou pohybovat}
 _putpixel(x1,y1,color);{?}
 if dyl>dxl then poml:=dyl shr 1
            else poml:=dxl shr 1;
 if dyp>dxp then pomp:=dyp shr 1
            else pomp:=dxp shr 1;
  repeat
  inc(y);
  {xxx obe cary posunout tak, aby yi=y a byl to nejvzdalenejsi mozny bod}
  _hline(xl,xp,y,color);
  until y=y2;

{if dx>=dy then begin
               pom:=dx shr 1;
                repeat
                inc(x1,krokx);
                inc(pom,dy);
                if pom>=dx then begin
                                dec(pom,dx);
                                inc(y1,kroky);
                                end;
                pixel;
                until x1=x2;
               end
          else begin
               pom:=dy shr 1;
                repeat
                inc(y1,kroky);
                inc(pom,dx);
                if pom>=dy then begin
                                dec(pom,dy);
                                inc(x1,krokx);
                                end;
                pixel;
                until y1=y2;
               end;}

 end;
if y2<>y3 then
 begin{dolni cast trojuhelnika}

 end;
End;{_filledtriangle}*)


procedure _FilledPolygon(var VertexArray;NumOfVertexes,color:byte);
type TVertex=record
             x,y:integer;
             end;
     TVertexArray=array[1..256] of TVertex;
Begin
if NumOfVertexes=1 then _PutPixel(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,color)
else if NumOfVertexes=2 then _Line(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
                                   TVertexArray(VertexArray)[2].x,TVertexArray(VertexArray)[2].y,color)
else
 repeat
 _FilledTriangle(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
                 TVertexArray(VertexArray)[NumOfVertexes-1].x,TVertexArray(VertexArray)[NumOfVertexes-1].y,
                 TVertexArray(VertexArray)[NumOfVertexes].x,TVertexArray(VertexArray)[NumOfVertexes].y,
                 color);
 Dec(NumOfVertexes);
 until NumOfVertexes=2;
End;{_filledpolygon}

(***************************************************************************
                      Kreslici procedury s orezavanim
****************************************************************************)

function ClipLine(var x1,y1,x2,y2:integer):boolean;
{Orizne souradnice cary podle okna. Vraci true, pokud z cary zbyl aspon
kousek, ktery by sel zobrazit.}
var otoceno:boolean;
Begin
otoceno:=false;
 repeat
 if (x1>=_minclipx)and(x1<=_maxclipx)and
    (x2>=_minclipx)and(x2<=_maxclipx)and
    (y1>=_minclipy)and(y1<=_maxclipy)and
    (y2>=_minclipy)and(y2<=_maxclipy) then begin {cara je urcite cela v okne => OK, muzeme kreslit}
                                           clipline:=true;
                                           break;
                                           end;
 if (x1<_minclipx)and(x2<_minclipx)or
    (x1>_maxclipx)and(x2>_maxclipx)or
    (y1<_minclipy)and(y2<_minclipy)or
    (y1>_maxclipy)and(y2>_maxclipy) then begin {cara je urcite cela mimo okno => OK, neni co kreslit}
                                         clipline:=false;
                                         break;
                                         end;
 {Existuji pripady, kdy predchozi podminka rekne, ze cara neni venku, ale
 pritom ve skutecnosti venku je. To ale nevadi - cara se trochu zkrati a v
 pristim cyklu uz bude jasne, ze je opravdu venku a skonci se.}
 if x1<_minclipx then begin {1 vycuhuje vlevo}
                      y1:=y1+longint(y2-y1)*(_MinClipX-x1)div(x2-x1);
                      {deleni nulou se nemusime bat, protoze kdyby ted bylo
                      x2=x1, musela by byt obe x mimo okno, coz by ale
                      odchytila predchozi podminka a uz bychom tady nebyli}
                      x1:=_MinClipX;
                      end
   else if y1<_minclipy then begin {1 vycuhuje nahore}
                             x1:=x1+longint(x2-x1)*(_MinClipY-y1)div(y2-y1);
                             y1:=_MinClipY;
                             end
    else if x1>_maxclipx then begin {1 vycuhuje vpravo}
                              y1:=y1+longint(y2-y1)*(_MaxClipX-x1)div(x2-x1);
                              x1:=_MaxClipX;
                              end
     else if y1>_maxclipy then begin {1 vycuhuje dole}
                               x1:=x1+longint(x2-x1)*(_MaxClipY-y1)div(y2-y1);
                               y1:=_MaxClipY;
                               end
      else begin
           {Bod 1 nevycuhuje, takze vycuhuje jenom bod 2. Abychom nemuseli
           psat dalsi ctyri podminky pro bod 2, radsi oba body prohodime.}
           xchange(x1,x2); xchange(y1,y2);
           otoceno:=not otoceno;
           end;
 {Mohli bychom napsat ty ctyri podminky tak, ze by se v jednom cyklu provedlo
 veskere orezavani najednou? Ne. Vypada to sice, ze bychom tak usetrili par
 pruchodu cyklem, ale ve skutecnosti by to nefungovalo. Mohlo by se totiz
 stat, ze by se cara zkratila podle x a hned potom zase protahla podle y a
 vysla by nekonecna smycka. Proto je nutne, aby se po kazdem "rezu" zkontrolo-
 valo, jestli uz cara neni venku z okna.}
 until false; {konci se Breakem}
if otoceno then begin {u maskovane cary je potreba, aby se kreslila od spravneho konce, proto tohle}
                xchange(x1,x2);
                xchange(y1,y2);
                end;
End;{clipline}

procedure __Line(x1,y1,x2,y2:integer;color:byte);
Begin
if ClipLine(x1,y1,x2,y2) then _Line(x1,y1,x2,y2,color);
End;{__line}

procedure __xorLine(x1,y1,x2,y2:integer;color:byte);
Begin
if ClipLine(x1,y1,x2,y2) then _xorLine(x1,y1,x2,y2,color);
End;{__xorline}

procedure __MaskedLine(x1,y1,x2,y2:integer;color:byte;Mask:word);
Begin
{zajisteni spravneho posunu masky:}
asm
mov CX,_minclipx
sub CX,x1
js @xNe         {kdyz x1 nevycuhuje, preskocime tohle}
 and CX,15      {posun bude o (_minclipx-x1) mod 16}
 rol mask,CL    {posuneme masku}
 jmp @konec     {a koncime, pripadne vycuhovani y1 nas nezajima}
@xNe:
mov CX,_minclipy       {to same pro y1}
sub CX,y1
js @konec
 and CX,15
 rol mask,CL
@konec:
end;
{Tohle resi jenom pripad, kdy zacatek cary vycuhuje skrz levy okraj a y1
zaroven neni nad hornim okrajem, nebo kdyz cara couha skrz horni okraj a x1
zaroven neni vlevo od leveho. Vycuhovani pravym a dolnim okrajem neni reseno
vubec, byly by to jen dalsi dve podminky.}
if ClipLine(x1,y1,x2,y2) then _MaskedLine(x1,y1,x2,y2,color,mask);
End;{__maskedline}

procedure __xorMaskedLine(x1,y1,x2,y2:integer;color:byte;Mask:word);
Begin
asm
mov CX,_minclipx
sub CX,x1
js @xNe
 and CX,15
 rol mask,CL
 jmp @konec
@xNe:
mov CX,_minclipy
sub CX,y1
js @konec
 and CX,15
 rol mask,CL
@konec:
end;
if ClipLine(x1,y1,x2,y2) then _xormaskedline(x1,y1,x2,y2,color,mask);
End;{__xormaskedline}

procedure __HLine(x1,x2,y:integer;color:byte);
Begin
if x1>x2 then asm  {kontrola poradi xovych souradnic}
              mov AX,x1
              mov BX,x2
              mov x1,BX
              mov x2,AX
              end;
if (y<_MinClipY)or(y>_MaxClipY)or(x1>_MaxClipX)or(x2<_MinClipX) then Exit;
if x1<_MinClipX then x1:=_MinClipX;
if x2>_MaxClipX then x2:=_MaxClipX;
_HLine(x1,x2,y,color);
End;{__hline}

procedure __XorHLine(x1,x2,y:integer;color:byte);
Begin
{$ifdef blbuvzdornost}
if x1>x2 then asm
              mov AX,x1
              mov BX,x2
              mov x1,BX
              mov x2,AX
              end;
{$endif}
if (y<_MinClipY)or(y>_MaxClipY)or(x1>_MaxClipX)or(x2<_MinClipX) then Exit;
if x1<_MinClipX then x1:=_MinClipX;
if x2>_MaxClipX then x2:=_MaxClipX;
_xorhline(x1,x2,y,color);
End;{__xorhline}

procedure __VLine(x,y1,y2:integer;color:byte);
Begin
{$ifdef blbuvzdornost}
if y1>y2 then asm
              mov AX,y1
              mov BX,y2
              mov y1,BX
              mov y2,AX
              end;
{$endif}
if (x<_MinClipX)or(x>_MaxClipX)or(y1>_MaxClipY)or(y2<_MinClipY) then Exit;
if y1<_MinClipY then y1:=_MinClipY;
if y2>_MaxClipY then y2:=_MaxClipY;
_VLine(x,y1,y2,color);
End;{__vline}

procedure __XorVLine(x,y1,y2:integer;color:byte);
Begin
{$ifdef blbuvzdornost}
if y1>y2 then asm
              mov AX,y1
              mov BX,y2
              mov y1,BX
              mov y2,AX
              end;
{$endif}
if (x<_MinClipX)or(x>_MaxClipX)or(y1>_MaxClipY)or(y2<_MinClipY) then Exit;
if y1<_MinClipY then y1:=_MinClipY;
if y2>_MaxClipY then y2:=_MaxClipY;
_xorvline(x,y1,y2,color);
End;{__xorvline}

procedure __Bar(x1,y1,x2,y2:integer;vbarva:byte);
Begin
{$ifdef blbuvzdornost}
asm
mov AX,x1
cmp AX,x2
js @xOK
 mov AX,x1
 mov BX,x2
 mov x1,BX
 mov x2,AX
@xOK:
mov AX,y1
cmp AX,y2
js @yOK
 mov AX,y1
 mov BX,y2
 mov y1,BX
 mov y2,AX
@yOK:
end;
{$endif}
if (x1>_MaxClipX)or(y1>_MaxClipY)or(x2<_MinClipX)or(y2<_MinClipY) then Exit;
if x1<_MinClipX then x1:=_MinClipX;
if y1<_MinClipY then y1:=_MinClipY;
if x2>_MaxClipX then x2:=_MaxClipX;
if y2>_MaxClipY then y2:=_MaxClipY;
_Bar(x1,y1,x2,y2,vbarva);
End;{__bar}

procedure __MaskedBar(x1,y1,x2,y2:integer;barva:word;var vzorek:_pole8bytu);
Begin
{$ifdef blbuvzdornost}
asm
mov AX,x1
cmp AX,x2
js @xOK
 mov AX,x1
 mov BX,x2
 mov x1,BX
 mov x2,AX
@xOK:
mov AX,y1
cmp AX,y2
js @yOK
 mov AX,y1
 mov BX,y2
 mov y1,BX
 mov y2,AX
@yOK:
end;
{$endif}
if (x1>_MaxClipX)or(y1>_MaxClipY)or(x2<_MinClipX)or(y2<_MinClipY) then Exit;
if x1<_MinClipX then x1:=_MinClipX;
if y1<_MinClipY then y1:=_MinClipY;
if x2>_MaxClipX then x2:=_MaxClipX;
if y2>_MaxClipY then y2:=_MaxClipY;
_maskedBar(x1,y1,x2,y2,barva,vzorek);
End;{__maskedbar}

procedure __Rectangle(x1,y1,x2,y2:word;color:byte);
Begin
__hline(x1,x2,y1,color);
__vline(x1,y1,y2,color);
__vline(x2,y1,y2,color);
__hline(x1,x2,y2,color);
End;{__rectangle}

procedure __XorRectangle(x1,y1,x2,y2:word;color:byte);
Begin
asm
mov AX,x1
cmp AX,x2
js @xOK
 mov AX,x1
 mov BX,x2              {jestli x1>x2, prohodime je}
 mov x1,BX
 mov x2,AX
@xOK:
mov AX,y1
cmp AX,y2
js @yOK
 mov AX,y1
 mov BX,y2              {jestli y1>y2, prohodime je}
 mov y1,BX
 mov y2,AX
@yOK:
end;
__xorhline(x1,x2,y1,color);
__xorvline(x1,y1,y2,color);
__xorvline(x2,y1,y2,color);
__xorhline(x1,x2,y2,color);
End;{__xorrectangle}

procedure __Box(x1,y1,x2,y2:integer;lhbarva,pdbarva,vbarva:byte);
Begin
{$ifdef blbuvzdornost}
asm
mov AX,x1
cmp AX,x2
js @xOK
 mov AX,x1
 mov BX,x2
 mov x1,BX
 mov x2,AX
@xOK:
mov AX,y1
cmp AX,y2
js @yOK
 mov AX,y1
 mov BX,y2
 mov y1,BX
 mov y2,AX
@yOK:
end;
{$endif}
__HLine(x1,x2,y1,lhbarva);
__VLine(x1,y1,y2,lhbarva);
__VLine(x2,y1,y2,pdbarva);
__HLine(x1,x2,y2,pdbarva);
if vbarva<>0 then
  begin
  if (x1>=_MaxClipX)or(y1>=_MaxClipY)or(x2<=_MinClipX)or(y2<=_MinClipY) then Exit;
  if x1<_MinClipX-1 then x1:=_MinClipX-1;
  if y1<_MinClipY-1 then y1:=_MinClipY-1;
  if x2>_MaxClipX+1 then x2:=_MaxClipX+1;
  if y2>_MaxClipY+1 then Y2:=_MaxClipY+1;
  _Bar(x1+1,y1+1,x2-1,y2-1,vbarva);
  end;
End;{__box}

procedure __MaskedBox(x1,y1,x2,y2:word;lhbarva,pdbarva:byte;vbarva:word;var vzorek:_pole8bytu);
Begin
{$ifdef blbuvzdornost}
asm
mov AX,x1
cmp AX,x2
js @xOK
 mov AX,x1
 mov BX,x2
 mov x1,BX
 mov x2,AX
@xOK:
mov AX,y1
cmp AX,y2
js @yOK
 mov AX,y1
 mov BX,y2
 mov y1,BX
 mov y2,AX
@yOK:
end;
{$endif}
__hline(x1,x2,y1,lhbarva);
__vline(x1,y1,y2,lhbarva);
__vline(x2,y1,y2,pdbarva);
__hline(x1,x2,y2,pdbarva);
if (vbarva<>0)and(x2-x1>=2)and(y2-y1>=2) then __maskedbar(x1+1,y1+1,x2-1,y2-1,vbarva,vzorek);
End;{__MaskedBox}

procedure __Ellipse(xc,yc:integer;rx,ry:word;color:byte);
var x,y:word;
    a,b,as,tas,bs,tbs:longint;
    d,dx,dy:longint;
Begin
if rx=0 then begin
             __VLine(xc,yc-ry,yc+ry,Color);
             Exit;
             end;
if ry=0 then begin
             __HLine(xc-rx,xc+rx,yc,Color);
             Exit;
             end;
x:=0; y:=ry; a:=rx; b:=ry; as:=a*a;
tas:=as shl 1; bs:=b*b; tbs:=bs shl 1;
d:=bs-as*b+(as shr 2); dx:=0; dy:=tas*b;
while dx<dy do begin
               __PutPixel(xc+x,yc+y,Color);
               __PutPixel(xc-x,yc+y,Color);
               __PutPixel(xc+x,yc-y,Color);
               __PutPixel(xc-x,yc-y,Color);
               if d>0 then begin y:=y-1; dy:=dy-tas; d:=d-dy; end;
               Inc(x); dx:=dx+tbs; d:=d+bs+dx;
               end;
d:=d+((3*(as-bs)div 2-(dx+dy))div 2);
while y>0 do begin
             __PutPixel(xc+x,yc+y,Color);
             __PutPixel(xc-x,yc+y,Color);
             __PutPixel(xc+x,yc-y,Color);
             __PutPixel(xc-x,yc-y,Color);
             if d<0 then begin Inc(x); dx:=dx+tbs; d:=d+dx; end;
             y:=y-1; dy:=dy-tas; d:=d+as-dy;
             end;
__PutPixel(xc+x,yc,Color);
__PutPixel(xc-x,yc,Color);
End;{__ellipse}

procedure __xorEllipse(xc,yc:integer;rx,ry:word;color:byte);
var x,y:word;
    a,b,as,tas,bs,tbs:longint;
    d,dx,dy:longint;
Begin
if rx=0 then begin
             __xorVLine(xc,yc-ry,yc+ry,Color);
             Exit;
             end;
if ry=0 then begin
             __xorHLine(xc-rx,xc+rx,yc,Color);
             Exit;
             end;
x:=0; y:=ry; a:=rx; b:=ry; as:=a*a;
tas:=as shl 1; bs:=b*b; tbs:=bs shl 1;
d:=bs-as*b+(as shr 2); dx:=0; dy:=tas*b;
while dx<dy do begin
               __xorPixel(xc+x,yc+y,Color);
               __xorPixel(xc-x,yc+y,Color);
               __xorPixel(xc+x,yc-y,Color);
               __xorPixel(xc-x,yc-y,Color);
               if d>0 then begin y:=y-1; dy:=dy-tas; d:=d-dy; end;
               Inc(x); dx:=dx+tbs; d:=d+bs+dx;
               end;
d:=d+((3*(as-bs)div 2-(dx+dy))div 2);
while y>0 do begin
             __xorPixel(xc+x,yc+y,Color);
             __xorPixel(xc-x,yc+y,Color);
             __xorPixel(xc+x,yc-y,Color);
             __xorPixel(xc-x,yc-y,Color);
             if d<0 then begin Inc(x); dx:=dx+tbs; d:=d+dx; end;
             y:=y-1; dy:=dy-tas; d:=d+as-dy;
             end;
__xorPixel(xc+x,yc,Color);
__xorPixel(xc-x,yc,Color);
End;{__xorellipse}

procedure __FilledEllipse(xc,yc:integer;rx,ry:word;Color:byte);
var x,y:word;
    a,b,as,tas,bs,tbs:longint;
    d,dx,dy:longint;
Begin
 if rx=0 then begin
  __VLine(xc,yc-ry,yc+ry,Color); Exit;
 end;
 if ry=0 then begin
  __HLine(xc-rx,xc+rx,yc,Color); Exit;
 end;
 x:=0; y:=ry; a:=rx; b:=ry; as:=a*a;
 tas:=as shl 1; bs:=b*b; tbs:=bs shl 1;
 d:=bs-as*b+(as shr 2); dx:=0; dy:=tas*b;
 while dx<dy do begin
  __HLine(xc-x,xc+x,yc+y,Color);
  __HLine(xc-x,xc+x,yc-y,Color);
  if d>0 then begin y:=y-1; dy:=dy-tas; d:=d-dy; end;
  Inc(x); dx:=dx+tbs; d:=d+bs+dx;
 end;
 d:=d+((3*(as-bs)div 2-(dx+dy))div 2);
 while y>0 do begin
  __HLine(xc-x,xc+x,yc+y,Color);
  __HLine(xc-x,xc+x,yc-y,Color);
  if d<0 then begin Inc(x); dx:=dx+tbs; d:=d+dx; end;
  y:=y-1; dy:=dy-tas; d:=d+as-dy;
 end;
 __HLine(xc-x,xc+x,yc,Color);
End;{__FilledEllipse}

procedure __Triangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
Begin
__Line(x1,y1,x2,y2,color);
__Line(x2,y2,x3,y3,color);
__Line(x3,y3,x1,y1,color);
End;{__triangle}

procedure __xorTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
Begin
__xorLine(x1,y1,x2,y2,color);
__xorLine(x2,y2,x3,y3,color);
__xorLine(x3,y3,x1,y1,color);
End;{__xortriangle}

procedure __FilledTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
var ymin,xmin,ymid,xmid,ymax,xmax,xmid2:longint;
    i:integer;
Begin
if y1>y2 then begin xchange(y1,y2); xchange(x1,x2); end;
if y1>y3 then begin xchange(y1,y3); xchange(x1,x3); end;
if y2>y3 then begin xchange(y2,y3); xchange(x2,x3); end;
ymin:=y1; xmin:=x1; ymid:=y2; xmid:=x2; ymax:=y3; xmax:=x3;
if ymax=ymin then begin
                  if x1>x2 then xchange(x1,x2);
                  if x1>x3 then xchange(x1,x3);
                  if x2>x3 then xchange(x2,x3);
                  __hline(x1,x3,y1,color);
                  exit;
                  end;
if (x1=x2)and(x2=x3) then begin
                          _vline(x1,y1,y3,color);
                          exit;
                          end;
xmid2:=(xmax-xmin)*(ymid-ymin) div (ymax-ymin)+xmin;
{if xmid2<xmid then xchange(integer(xmid2),integer(xmid)); tohle nestaci}
if ymin<>ymid then for i:=ymin to ymid do __HLine((xmid-xmin)*(i-ymin) div (ymid-ymin)+xmin,
                                                  (xmid2-xmin)*(i-ymin) div (ymid-ymin)+xmin,i,color);
if ymid<>ymax then for i:=ymid to ymax do __HLine((xmax-xmid)*(i-ymid) div (ymax-ymid)+xmid,
                                                  (xmax-xmid2)*(i-ymid) div (ymax-ymid)+xmid2,i,color);
End;{__filledtriangle}

procedure __FilledPolygon(var VertexArray;NumOfVertexes,color:byte);
type TVertex=record
             x,y:integer;
             end;
     TVertexArray=array[1..256] of TVertex;
Begin
if NumOfVertexes=1 then __PutPixel(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,color)
else if NumOfVertexes=2 then __Line(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
                                    TVertexArray(VertexArray)[2].x,TVertexArray(VertexArray)[2].y,color)
else
 repeat
 __FilledTriangle(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
                  TVertexArray(VertexArray)[NumOfVertexes-1].x,TVertexArray(VertexArray)[NumOfVertexes-1].y,
                  TVertexArray(VertexArray)[NumOfVertexes].x,TVertexArray(VertexArray)[NumOfVertexes].y,color);
 Dec(NumOfVertexes);
 until NumOfVertexes=2;
End;{__filledpolygon}

procedure __Fill(color:byte);
Begin
__bar(_minclipx,_minclipy,_maxclipx,_maxclipy,color);
End;{__fill}

procedure __FloodFill(ffx,ffy:integer; barva:byte);
const MaxZ=8;{maximalni pocet zasobniku}
type polepixelu=array[0..0] of record
                               x,y:word;
                               end;
     uknapolepixelu=^polepixelu;
var pix:byte;{pro ulozeni barvy pocatecniho pixelu, pres kterou se nova barva rozlije}
    zasobniky:array[1..maxz] of record
                                data:uknapolepixelu;
                                velikost:word;{velikost zasobniku [B]}
                                max:word;{maximalni pouzitelny index}
                                vrchol:word;{index volne pozice na vrcholu zasobniku}
                                end;
    poslednizasobnik:byte;{cislo prave pouzivaneho zasobniku}
    i,j:integer;
{}function NovyZasobnik:boolean;{vytvori novy zasobnik a vraci true, jestli se to povedlo}
{}Begin
{}novyzasobnik:=false;
{}if poslednizasobnik<maxz then
{} with zasobniky[poslednizasobnik+1] do
{}   begin
{}   velikost:=65520;
{}   if velikost>maxavail then velikost:=maxavail;
{}   if velikost<16 then exit;{uz neni pamet}
{}   max:=(velikost shr 2)-1;{deleno velikosti jednoho prvku pole plus jeden prvek rezerva kvuli indexovani pole od nuly}
{}   vrchol:=0;
{}   getmem(data,velikost);
{}   inc(poslednizasobnik);
{}   novyzasobnik:=true;
{}   end;
{}End;{novyzasobnik}
{}procedure _Push(_x,_y:integer);
{}Begin
{}if ((poslednizasobnik=0)or(zasobniky[poslednizasobnik].vrchol>zasobniky[poslednizasobnik].max))
{}  and (not novyzasobnik) then exit;
{}with zasobniky[poslednizasobnik] do begin
{}                                    data^[vrchol].x:=_x;
{}                                    data^[vrchol].y:=_y;
{}                                    inc(vrchol);
{}                                    end;
{}End;{_push}
{}procedure ZrusZasobnik;
{}Begin
{}if poslednizasobnik<>0 then
{} begin
{} with zasobniky[poslednizasobnik] do freemem(data,velikost);
{} dec(poslednizasobnik);
{} end;
{}End;{zruszasobnik}
{}procedure _Pop(var _x,_y:integer);
{}Begin
{}_x:=ffx; _y:=ffy;{pro pripad problemu}
{}if poslednizasobnik=0 then exit;
{}if zasobniky[poslednizasobnik].vrchol=0 then zruszasobnik;
{}if poslednizasobnik=0 then exit;
{}with zasobniky[poslednizasobnik] do begin
{}                                    dec(vrchol);
{}                                    _x:=data^[vrchol].x;
{}                                    _y:=data^[vrchol].y;
{}                                    end;
{}End;{_pop}
Begin{__floodfill}
if (ffx>_maxclipx)or(ffy>_maxclipy)or(ffx<_minclipx)or(ffy<_minclipy) then exit;{jsme mimo obrazovku}
poslednizasobnik:=0;
pix:=_getpixel(ffx,ffy);
if pix=barva then exit;{neni co barvit}
_push(ffx,ffy);
 repeat
 _pop(j,i);{precteme souradnice pixelu}
 _putpixel(j,i,barva);{vybarvime ho}
 {pixel nahore (pokud se nad timhle nachazi pixel, ktery se ma vybarvit, ulozime ho):}
 if (i>_minclipy)and (_getpixel(j,i-1)=pix) then _push(j,i-1);
 {pixel dole (to same):}
 if (i<_maxclipy)and (_getpixel(j,i+1)=pix) then _push(j,i+1);
 {pixel vlevo:}
 if (j>_minclipx)and (_getpixel(j-1,i)=pix) then _push(j-1,i);
 {pixel vpravo:}
 if (j<_maxclipx)and (_getpixel(j+1,i)=pix) then _push(j+1,i);
 until poslednizasobnik=0;{koncime, kdyz je prazdny zasobnik}
End;{__floodfill}

(****************************************************************************
                                pro texty:
*****************************************************************************)

procedure _NastavFont(fn:_font;loaddefault:boolean);
Begin
__font:=fn.fdata;
__vyska:=fn.fvyska;
if loaddefault then with fn do _nastavtext(fstdx,fstdy,fstdroztec);
End;{_nastavfont}

procedure _NastavText(velikostx,velikosty,roztec:byte);
Begin
if velikostx<>0 then __velikostx:=velikostx;
if velikosty<>0 then __velikosty:=velikosty;
if roztec<>0 then __roztec:=roztec;
End;{_nastavtext}

{pro _Pushtext a _Poptext:}
type
_uknatxtsettings=^_txtsettings;
_txtsettings=record
             vyska:byte;
             font:_uknafont;
             velikostx,
             velikosty,
             roztec:byte;
             dalsi:_uknatxtsettings;
             end;
const
_prvnits:_uknatxtsettings=nil;

procedure _pushtext;
var n:_uknatxtsettings;
Begin
new(n);
with n^ do begin
           vyska:=__vyska;
           font:=__font;
           velikostx:=__velikostx;
           velikosty:=__velikosty;
           roztec:=__roztec;
           end;
if _prvnits=nil then begin
                     n^.dalsi:=nil;
                     _prvnits:=n;
                     end
                else begin
                     n^.dalsi:=_prvnits;
                     _prvnits:=n;
                     end;
End;{_pushtext}

procedure _poptext;
var v:_uknatxtsettings;
Begin
if _prvnits=nil then exit;
with _prvnits^ do begin
                  __vyska:=vyska;
                  __font:=font;
                  __velikostx:=velikostx;
                  __velikosty:=velikosty;
                  __roztec:=roztec;
                  end;
v:=_prvnits;
_prvnits:=_prvnits^.dalsi;
dispose(v);
End;{_poptext}

procedure _lodtext;
Begin
with _prvnits^ do begin
                  __vyska:=vyska;
                  __font:=font;
                  __velikostx:=velikostx;
                  __velikosty:=velikosty;
                  __roztec:=roztec;
                  end;
End;{_lodtext}

procedure _NactiFont(var FN:_font; var Soubor:file);
var velikost:word;
    b:byte;
Begin
fn.fdata:=nil; {jestli se neco pokazi, tohle tady zustane}
blockread(soubor,fn.fvyska,1);
velikost:=256*fn.fvyska; {je 256 znaku a kazdy ma tolik bytu, jak je vysoky}
if (ioresult<>0)or(velikost>maxavail) then exit;
getmem(fn.fdata,velikost);
blockread(soubor,fn.fstdroztec,1);
blockread(soubor,b,1);
fn.fstdx:=b and 15;
fn.fstdy:=b shr 4;
blockread(soubor,fn.fdata^,velikost);
if ioresult<>0 then begin
                    freemem(fn.fdata,velikost);{kdyz se font nenacetl, tak nema cenu zbytecne zabirat pamet}
                    fn.fdata:=nil;
                    end;
End;{_nactifont}

procedure _NactiFont2(var FN:_font; JmenoSouboru:string);
var f:file;
Begin
assign(f,jmenosouboru);
reset(f,1);
_nactifont(fn,f);
close(f);
if ioresult<>0 then ;{reset Ioresultu pro pripad chyby pri Close (i to se muze stat)}
End;{_nactifont2}

function _UlozFont(FN:_font;soubor:string):boolean;
var f:file;
    b:byte;
Begin
_ulozfont:=false;
assign(f,soubor);
rewrite(f,1);
if ioresult=0 then begin
                   blockwrite(f,fn.fvyska,1);
                   blockwrite(f,fn.fstdroztec,1);
                   b:=(fn.fstdy shl 4) or fn.fstdx;
                   blockwrite(f,b,1);
                   blockwrite(f,fn.fdata^,256*fn.fvyska);
                   close(f);
                   _ulozfont:=(ioresult=0);
                   end
              else _ulozfont:=false;
End;{_ulozfont}

procedure _ZrusFont(var fn:_font);
Begin
if fn.fdata<>nil then begin
                      freemem(fn.fdata,fn.fvyska*256);
                      fn.fdata:=nil; fn.fvyska:=0;{aby bylo jasne, ze uz tu nic neni}
                      end;
End;{_zrusfont}

procedure _print(ix,iy:integer;barva:byte;textik:string);
var PoziceVRetezci,{index pro pohyb mezi znaky retezce}
    CisloRadku,{index pro pohyb mezi radky pismena}
    IndexY,{index pro kresleni jednotlivych kopii jednoho radku pismena (pro _velikostY>1)}
    radek:byte;{data - jeden radek pismena}
    y:integer;{pro vypocet a ulozeni skutecne yove souradnice radku na obrazovce}
    i:word;{ridici promenna do asm cyklu}
    kod:integer;{pro Val}
    znak:word;{pomocna, pro zadavani znaku pres ordinalni cisla (%%z...)}
Begin
if (__font=nil)or(ix+_sirkatextu(textik)<_minclipx)
 or(ix+__velikostx shl 3>=_maxclipx)or(iy>_maxclipy)
 or(iy+_vyskatextu<_minclipy) then exit;
PoziceVRetezci:=1;
while ix<_minclipx do begin
                      inc(ix,__roztec);
                      inc(PoziceVRetezci);
                      end;
if PoziceVRetezci>byte(textik[0]) then exit;
 repeat{cyklus pro kazdy znak retezce}
 {osetreni ridicich kodu:}
 znak:=$FFFF;
 if (textik[pozicevretezci]='%') then
 {nastaveni barvy "za letu":}
  if (copy(textik,pozicevretezci+1,2)='%b')and(pozicevretezci<byte(textik[0])-4)
    then begin
         val(copy(textik,pozicevretezci+3,3),i,kod);{prevod na cislo}
         if kod=0 then begin{prevod se povedl}
                       if _ridicikody=1 then barva:=i;{nastavime barvu...}
                       if _ridicikody<>2 then inc(pozicevretezci,6);{...a presuneme se za ridici sekvenci}
                       end;
         end
   {vlozeni znaku pomoci ordinalniho cisla:}
   else if (copy(textik,pozicevretezci+1,2)='%z')and(pozicevretezci<byte(textik[0])-4)
     then begin
          val(copy(textik,pozicevretezci+3,3),i,kod);{prevod na cislo}
          if kod=0 then begin{prevod se povedl}
                        if _ridicikody=1 then znak:=i;{nastavime znak...}
                        if _ridicikody<>2 then inc(pozicevretezci,5);{...a presuneme se za ridici sekvenci}
                        end;
          end;
 for CisloRadku:=0 to __vyska-1 do {pro kazdy radek znaku}
  begin
  if znak=$FFFF then radek:=__font^[__vyska*byte(textik[PoziceVRetezci])+cisloradku]{nalezeni prislusneho bytu ve fontu}
                else radek:=__font^[__vyska*znak+cisloradku];
  for indexY:=0 to __velikostY-1 do {pro kazdou kopii radku pismena zobrazenou na obrazovce}
   begin
   y:=iy+cisloradku*__velikostY+indexY; {skutecna poloha zacatku radku}
   if (y>=_minclipy)and(y<=_maxclipy)then {kdyz je radek na obrazovce}
    begin
    i:=8;       {sirka pismena (konst.)}
    {vykresleni radku:}
    if __velikostx<=3 then
    asm{kresli se pixel po pixelu (low-level)}
    mov AX,__lhy
    add y,AX      {pro virtualni obrazovku}
    mov AX,$A000
    mov ES,AX      {ES = segment pro videopamet}

    xor CH,CH     {bude potreba jen CL}
    xor BX,BX     {pro __switchbank}

    mov AX,_MaxX
    inc AX
    mul y
    add AX,ix
    adc DX,0       {DX = banka}
    mov DI,AX      {DI = ofset (oboje pro prvni pixel radku)}

    cmp DX,CurBank
    je @neprepinej1
     mov curbank,DX                         {...rozepsany Bankswitch...}
     mov AX,DX
     mul Granularity
     mov DX,AX
     mov AX,$4F05
     call __switchbank
    @neprepinej1:          {banka nastavena, jeji cislo je stale v DX}

     @ProKazdyPixelRadku:     {cyklus for i:=8 downto 1}
     mov CL,__velikostX   {CX = pocet zobrazenych pixelu na jeden pixel bitmapy znaku}
     rol radek,1        {je ve zdrojovem znaku v prislusnem bitu 1?}
     jc @PixelBude
      {jinak vynechani pixelu:}
      add DI,CX        {posun o sirku pixelu dal}
      jnc @neprepinej2 {kdyz DI nepretece, banka se neprepina}
       inc DX
       mov curbank,DX
       mov AX,DX
       mul Granularity
       mov DX,AX
       mov AX,$4F05
       call __switchbank
      @neprepinej2:          {banka nastavena, jeji cislo je stale v DX}
      jmp @hotovo
     @PixelBude:
      {nakresleni pixelu:}
       @ProKazdyPixelPixelu:   {cyklus for CX:=__velikostX downto 1}
       mov AL,barva
       mov ES:[DI],AL       {vykresleni pixelu}
       inc DI               {o pixel na obrazovce dal}
       jnz @neprepinej3     {kdyz DI pretece (dalsi banka), bude ZF = 1, CF se nenastavi}
        inc DX
        mov curbank,DX
        mov AX,DX
        mul Granularity
        mov DX,AX
        mov AX,$4F05
        call __switchbank
       @neprepinej3:          {banka nastavena, jeji cislo je stale v DX}
       loop @ProKazdyPixelPixelu
     @hotovo:
     dec i
     jnz @ProKazdyPixelRadku
    end
    else
    asm {kresli se pomoci 32bitove _Hline, protoze je to tak rychlejsi}
    @proKazdyPixelRadku:
    rol radek,1
    jnc @dira
     mov AX,8
     sub AX,i
     mul __velikostx
     add AX,ix
     push AX   {x1}
     add AL,__velikostx
     adc AH,0
     dec AX
     push AX   {x2}
     push y    {y}
     push word ptr barva {barva}
     call _hline      {kresleni}
    @dira:
    dec i
    jnz @ProKazdyPixelRadku
    end;
    end;
   end;
  end;
 inc(ix,__roztec);
 inc(pozicevretezci);
 until (pozicevretezci>byte(textik[0]))or(ix>=_maxclipx-__velikostx shl 3);
End;{_print}

function _SirkaTextu(retezec:string):word;
Begin
if __roztec>8 then if byte(retezec[0])=0 then _sirkatextu:=0
                                         else _sirkatextu:=(byte(retezec[0])-1)*__roztec+__velikostx shl 3
              else _sirkatextu:=byte(retezec[0])*__roztec;
End;{_sirkatextu}

function _SirkaTextu2(PocetZnaku:byte):word;
Begin
if __roztec>8 then if pocetznaku=0 then _sirkatextu2:=0
                                   else _sirkatextu2:=(pocetznaku-1)*__roztec+__velikostx shl 3
              else _sirkatextu2:=pocetznaku*__roztec;
End;{_sirkatextu2}

function _VyskaTextu:word;
Begin
_vyskatextu:=__velikosty*__vyska;{= velikost pixelu * vyska pisma v pixelech}
End;{_vyskatextu}

(***************************************************************************
                       Procedury pro praci s obrazky
****************************************************************************)

procedure _PutTImage(x1,y1:integer;width,height:word;clip:pointer);
var i,SourcSegment,SourcOfset:word;
    xshift,yshift,dssave,width2,incr:word;
Begin
if (x1>_MaxX)or(y1>_MaxY) then Exit;
xshift:=0; yshift:=0;
if x1<0 then begin{vycuhuje levy okraj?}
             if x1+width<=0 then Exit;
             xshift:=-x1;
             width:=width+x1;
             x1:=0;
             end;
if y1<0 then begin{vycuhuje horni okraj?}
             if y1+height<=0 then Exit;
             yshift:=-y1;
             height:=height+y1;
             y1:=0;
             end;
sourcsegment:=seg(clip^);
SourcOfset:=ofs(clip^)+yshift*(width+xshift)+xshift;
if (x1+width-1)>_MaxX then begin{vycuhuje pravy okraj?}
                           inc(xshift,x1+width-1-_MaxX);
                           width:=_MaxX-x1+1;
                           end;
if (y1+height-1)>_MaxY then height:=_MaxY-y1+1;{vycuhuje dolni okraj?}
incr:=_maxx-width+1;
i:=y1+__lhy;
inc(height,i);{height uz neni vyska, ale yova souradnice posledniho radku +1}
{nahrazuje to cyklus for i:=y1 to y1+height-1 do...}
asm
mov AX,DS
mov dssave,AX   {ulozeni DS (myslim, ze to s pomocnou promennou bude trochu rychlejsi nez s push+pop)}

mov AX,$A000
mov ES,AX       {ES = cilovy segment}

mov AX,sourcofset
mov SI,AX       {SI = zdrojovy ofset (zacatek prvniho radku)}

mov AX,_Maxx
inc AX
mul i         {AX = ofset zacatku radku na obrazovce, DX = banka na tom zacatku}

add AX,x1    {AX = ofset bodu, kde bude zacinat radek obrazku}
adc DX,0     {DX = banka toho bodu}
mov DI,AX    {DI = cilovy ofset}

{ted je v DI ofset a v DX banka}

@zacatek: {zacatek cyklu, ktery probehne pro kazdy radek}

mov BX,DI
mov CX,DX
add BX,width
adc CX,0
dec BX         {BX = ofset bodu na konci radku, CX = banka toho bodu}

{ted je v DI ofset zacatku radku, v DX jeho banka a v BX a CX to same pro konec radku}

cmp DX,curbank  {prepnuti do banky pro zacatek radku}
je @neprepinej
 mov AX,DX
 mov curbank,AX
 mul granularity
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank
@neprepinej:

cmp CX,curbank {je banka na konci radku stejna jako na zacatku?}
je @najednou  {je => hop dal ***}
 {neni => nakresli se prvni cast radku v prvni bance, pak se prepne o banku dal a nakresli se zbytek:}

 or CX,$FFFF  {= mov CX,$FFFF, ale trochu rychlejsi}
 sub CX,DI
 inc CX        {CX = pocet pixelu v prvni bance}

 mov AX,width
 sub AX,CX
 mov width2,AX   {width2 = pocet pixelu ve druhe bance}

 mov AX,sourcsegment
 mov DS,AX     {DS = zdrojovy segment}

  @PrvniPulkaRadku:
  lodsb            {zdroj -> AL}
  or AL,AL     {je AL 0 (pruhledna)?}
  jz @nic1
   mov ES:[DI],AL      {AL -> obrazovka}
  @nic1:
  inc DI        {v cili o pixel dal (i kdyz se nic nekreslilo)}
  loop @PrvniPulkaRadku
 {v posledni iteraci DI preteklo a je v nem 0, coz presne chceme}

 mov AX,dssave
 mov DS,AX  {obnoveni DS}

 mov AX,curbank
 inc AX                 {o banku dal}
 mov CurBank,AX
 mul Granularity
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank

 mov CX,width2  {CX = pocet pixelu ve druhe bance}
 jmp @jedeme

@najednou:     {***}
mov CX,width  {CX = pocet pixelu}

@jedeme:

or CX,CX
jz @nic      {aby se cyklus neprovedl vubec, pokud je CX=0}

mov AX,sourcsegment
mov DS,AX     {DS = opet zdrojovy segment}

 @ZbytekRadku:
 lodsb            {zdroj -> AL}
 or AL,AL
 jz @nic2
  mov ES:[DI],AL      {AL -> obrazovka}
 @nic2:
 inc DI
 loop @ZbytekRadku

mov AX,dssave
mov DS,AX      {obnoveni DS}

@nic:

{radek je vykreslen}

add SI,xshift  {ve zdroji o radek dal}

mov DX,curbank
add DI,incr      {na obrazovce o radek dal}
adc DX,0

{ted je v DI ofset a v DX banka, presne tak, jak to je na zacatku cyklu potreba}

inc i
mov AX,i
cmp AX,height    {uz jsme nakreslili posledni radek?}
jne @zacatek  {ne => jdeme na dalsi}
end;
End;{_puttimage}

procedure _PutRTImage(x1,y1:integer;width,height:word;clip:pointer);
{trochu slozitejsi Puttimage - ze zdroje se musi cist obracene}
var i,SourcSegment,SourcOfset,xshift,yshift,dssave,width2,incr:word;
Begin
if (x1>_MaxX)or(y1>_MaxY) then Exit;
xshift:=0; yshift:=0;
if x1<0 then begin{vycuhuje levy okraj?}
             if x1+width<=0 then Exit;
             xshift:=-x1;
             width:=width+x1;
             x1:=0;
             end;
if y1<0 then begin{vycuhuje horni okraj?}
             if y1+height<=0 then Exit;
             yshift:=-y1;
             height:=height+y1;
             y1:=0;
             end;
sourcsegment:=seg(clip^);
SourcOfset:=ofs(clip^)+width+yshift*(width+xshift);
if (x1+width-1)>_MaxX then begin{vycuhuje pravy okraj?}
                           inc(xshift,x1+width-1-_MaxX);
                           width:=_MaxX-x1+1;
                           end;
if (y1+height-1)>_MaxY then height:=_MaxY-y1+1;{vycuhuje dolni okraj?}
incr:=_maxx-width+1;
i:=y1+__lhy;
inc(height,i);{height uz neni vyska, ale yova souradnice posledniho radku +1}
{nahrazuje to cyklus for i:=y1 to y1+height-1 do...}
asm
mov AX,DS
mov dssave,AX   {ulozeni DS}

mov AX,$A000
mov ES,AX       {ES = cilovy segment}

mov AX,sourcofset
mov SI,AX       {SI = zdrojovy ofset (bod za koncem prvniho radku)}

mov AX,_Maxx
inc AX
mul i         {AX = ofset zacatku radku obrazovky, DX = banka na tom zacatku}

add AX,x1    {AX = ofset bodu, kde bude zacinat radek obrazku}
adc DX,0     {DX = banka toho bodu}
mov DI,AX    {DI = cilovy ofset}

{ted je v DI ofset a v DX banka}

@zacatek: {zacatek cyklu, ktery probehne pro kazdy radek}

mov BX,DI
mov CX,DX
add BX,width
adc CX,0
dec BX         {BX = ofset bodu na konci radku, CX = banka toho bodu}

{ted je v DI ofset zacatku radku, v DX jeho banka a v BX a CX to same pro konec radku}

cmp DX,curbank  {prepnuti do banky pro zacatek radku}
je @neprepinej
 mov AX,DX
 mov CurBank,AX
 mul Granularity
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank
@neprepinej:

cmp CX,curbank {je banka na konci radku stejna jako na zacatku?}
je @najednou  {je => hop dal ***}
 {neni => nakresli se prvni cast radku v prvni bance, pak se prepne o banku dal a nakresli se zbytek:}

 or CX,$FFFF  {mov CX,$FFFF}
 sub CX,DI
 inc CX        {CX = pocet pixelu v prvni bance}

 mov AX,width
 sub AX,CX
 mov width2,AX   {width2 = pocet pixelu ve druhe bance}

 mov AX,sourcsegment
 mov DS,AX     {DS = zdrojovy segment}

  @PrvniPulkaRadku:
  mov AL,DS:[SI-1]     {zdroj -> AL}
  dec SI     {slo by to i pres std a cld, ale takhle mi to pripada jednodussi}
  or AL,AL     {je AL 0 (pruhledna)?}
  jz @nic1
   mov ES:[DI],AL      {AL -> obrazovka}
  @nic1:
  inc DI        {v cili o pixel dal (i kdyz se nic nekreslilo)}
  loop @PrvniPulkaRadku
 {v posledni iteraci DI preteklo a je v nem 0, coz presne chceme}

 mov AX,dssave
 mov DS,AX  {obnoveni DS}

 mov AX,curbank
 inc AX                {prepnuti o banku dal (DI je 0, takze budeme kreslit na zacatek dalsi banky)}
 mov CurBank,AX
 mul Granularity
 mov DX,AX
 mov AX,$4F05
 xor BX,BX
 call __switchbank

 mov CX,width2  {CX = pocet pixelu ve druhe bance}
 jmp @jedeme

@najednou:
mov CX,width  {CX = pocet pixelu}

@jedeme:

or CX,CX
jz @nic

mov AX,sourcsegment
mov DS,AX     {DS = opet zdrojovy segment}

 @ZbytekRadku:
 mov AL,DS:[SI-1]   {zdroj -> AL}
 dec SI
 or AL,AL
 jz @nic2
  mov ES:[DI],AL      {AL -> obrazovka}
 @nic2:
 inc DI
 loop @ZbytekRadku

{SI je ted na zacatku radku ve zdroji}

mov AX,dssave
mov DS,AX      {obnoveni DS}

@nic:
{radek je vykreslen}

add SI,xshift      {ve zdroji o radek dal}
add SI,width
add SI,width       {(ze zacatku jednoho radku za konec dalsiho radku)}

mov DX,curbank
add DI,incr         {na obrazovce o radek dal}
adc DX,0
{ted je v DI ofset a v DX banka, presne tak, jak to je na zacatku cyklu potreba}

inc i
mov AX,i
cmp AX,height    {uz jsme nakreslili posledni radek?}
jne @zacatek  {ne => jdeme na dalsi}
end;
End;{_putrtimage}

procedure _PutImage(x1,y1:integer;width,height:word;clip:pointer);
{Podobna Puttimage, lisi se jen kopirovacim cyklem: misto pixel po pixelu s
testovanim jeho hodnoty se kopiruje po qwordech, dwordech a bytech, aby to
bylo rychlejsi.}
var xshift,yshift,sourcsegment,sourcofset,incr,i,DSsave,width2:word;
Begin
if (x1>_MaxX)or(y1>_MaxY) then Exit;
xshift:=0; yshift:=0;
if x1<0 then begin
             if x1+width<=0 then Exit;
             xshift:=-x1;
             width:=width+x1;
             x1:=0;
             end;
if y1<0 then begin
             if y1+height<=0 then Exit;
             yshift:=-y1;
             height:=height+y1;
             y1:=0;
             end;
sourcsegment:=seg(clip^);
SourcOfset:=ofs(clip^)+yshift*(width+xshift)+xshift;
if (x1+width-1)>_MaxX then begin
                           inc(xshift,x1+width-1-_MaxX);
                           width:=_MaxX-x1+1;
                           end;
if (y1+height-1)>_MaxY then height:=_MaxY-y1+1;
incr:=_maxx-width+1;
i:=y1+__lhy;
inc(height,i);
asm
mov AX,DS
mov dssave,AX
mov AX,$A000
mov ES,AX
mov AX,sourcofset
mov SI,AX
mov AX,_Maxx
inc AX
mul i
add AX,x1
adc DX,0
mov DI,AX
 @zacatek:
 mov BX,DI
 mov CX,DX
 add BX,width
 adc CX,0
 dec BX
 sbb CX,0
 cmp DX,curbank
 je @neprepinej
  mov AX,DX
  mov CurBank,AX
  mul Granularity
  mov DX,AX
  mov AX,$4F05
  xor BX,BX
  call __switchbank
 @neprepinej:
 cmp CX,curbank
 je @najednou
  or CX,$FFFF
  sub CX,DI
  inc CX
  mov AX,width
  sub AX,CX
  mov width2,AX
  mov AX,sourcsegment
  mov DS,AX
   {presun po qwordech:}
   mov BX,CX
   shr CX,4
   jz @MeneNez16_1
    @smycka1:
    fild qword ptr DS:[SI]           {komentare - viz jednotku NFSUP}
    fild qword ptr DS:[SI+8]
    fistp qword ptr ES:[DI+8]
    fistp qword ptr ES:[DI]
    add SI,16
    add DI,16
    loop @smycka1
   @MeneNez16_1:
   {presun po dwordech:}
   mov CX,BX
   and CX,15
   mov BX,CX
   shr CX,2
    db $66; rep movsw
   {presun po bytech:}
   mov CX,BX
   and CX,3
    rep movsb
  mov AX,dssave
  mov DS,AX
  mov AX,curbank
  inc AX
  mov CurBank,AX
  mul Granularity
  mov DX,AX
  mov AX,$4F05
  xor BX,BX
  call __switchbank
  mov CX,width2
  jmp @jedeme
 @najednou:
 mov CX,width
 @jedeme:
 mov AX,sourcsegment
 mov DS,AX
   mov BX,CX
   shr CX,4
   jz @MeneNez16_2
    @smycka2:
    fild qword ptr DS:[SI]
    fild qword ptr DS:[SI+8]
    fistp qword ptr ES:[DI+8]
    fistp qword ptr ES:[DI]
    add SI,16
    add DI,16
    loop @smycka2
   @MeneNez16_2:
   mov CX,BX
   and CX,15
   mov BX,CX
   shr CX,2
    db $66; rep movsw
   mov CX,BX
   and CX,3
    rep movsb
 mov AX,dssave
 mov DS,AX
 add SI,xshift
 mov DX,curbank
 add DI,incr
 adc DX,0
 inc i
 mov AX,i
 cmp AX,height
 jne @zacatek
end;
End;{_putimage}

procedure _GetImage(x1,y1,x2,y2:word;clip:pointer); assembler;
{Podobna Putimage, ale s prohozenym zdrojem a cilem. Jeste zjednodusena tim,
ze se nestara o vycuhovani obrazku z obrazovky a ze je jenom 32bitova.}
var i,dssave,width,width2,incr:word;
Asm
mov AX,__lhy
add y1,AX
add y2,AX
mov AX,y1
mov i,AX
inc y2
les DI,clip
mov AX,x2
sub AX,x1
inc AX
mov width,AX
mov AX,_maxx
sub AX,width
inc AX
mov incr,AX
mov AX,DS
mov dssave,AX
mov AX,_Maxx
inc AX
mul i
add AX,x1
adc DX,0
mov SI,AX
@zacatek:
mov BX,SI
mov CX,DX
add BX,width
adc CX,0
dec BX
sbb CX,0
cmp DX,curbank
je @neprepinej
 push DX
 call _SwitchBank
@neprepinej:
cmp CX,curbank
je @najednou
 or CX,$FFFF
 sub CX,SI
 inc CX
 mov AX,width
 sub AX,CX
 mov width2,AX
 mov AX,$A000
 mov DS,AX
  mov BX,CX
  shr CX,2
  db $66; rep movsw
  and BX,3
  mov CX,BX
  rep movsb
 mov AX,dssave
 mov DS,AX
 mov AX,curbank
 inc AX
 push AX
 call _switchbank
 mov CX,width2
 jmp @jedeme
@najednou:
 mov CX,width
 @jedeme:
 mov AX,$A000
 mov DS,AX
  mov BX,CX
  shr CX,2
  db $66; rep movsw
  and BX,3
  mov CX,BX
  rep movsb
 mov AX,dssave
 mov DS,AX
mov DX,curbank
add SI,incr
adc DX,0
inc i
mov AX,i
cmp AX,y2
jne @zacatek
End;{_getimage}

(***************************************************************************
                                   Ruzne
****************************************************************************)

procedure _VSync; assembler;
Asm
mov DX,$3DA
 @repeat:
 in AL,DX
 test AL,8
 jz @repeat
End;{_vsync}

procedure _VSyncStart; assembler;
Asm
mov DX,$3DA
 @repeat1:
 in AL,DX
 test AL,8
 jnz @repeat1
 @repeat2:
 in AL,DX
 test AL,8
 jz @repeat2
End;{_vsyncstart}

procedure _SetWindow(minx,miny,maxx,maxy:integer);
Begin
if minx<0 then _minclipx:=0
          else _minclipx:=minx;
if miny<0 then _minclipy:=0
          else _minclipy:=miny;
if maxx>_maxx then _maxclipx:=_maxx
              else _maxclipx:=maxx;
if maxy>_maxy then _maxclipy:=_maxy
              else _maxclipy:=maxy;
End;{_setwindow}

procedure _SetDefaultWindow;
Begin
_MinClipX:=0;
_MinClipY:=0;
_MaxClipX:=_MaxX;
_MaxClipY:=_MaxY;
End;{_setdefaultwindow}

{$ifdef povolshowbanks}
procedure _ShowBanks; assembler; {jen modifikovana _Fill}
var nbanks,bank,rest:word;
Asm
mov bx,_MaxX
inc bx
mov ax,_MaxY
inc ax
mul bx
mov nbanks,dx
mov rest,ax
mov ax,$A000
mov es,ax
xor ax,ax
mov bank,ax
@@1:
 push bank
 call SwitchBank
 xor di,di
 mov cx,$4000
 mov al,byte ptr bank
 mov ah,al
 mov bx,ax
 db $66; shl ax,16
 mov ax,bx
 db $66; rep stosw
 inc bank
 mov ax,bank
 dec ax
 cmp ax,nbanks
 jne @@1
push bank
call SwitchBank
xor di,di
mov cx,rest
inc cx
mov al,byte ptr bank
mov ah,al
mov bx,ax
db $66; shl ax,16
mov ax,bx
mov bx,cx
shr cx,2
db $66; rep stosw
and bx,3
mov cx,bx
rep stosb
End;{_showbanks}
{$endif}

(****************************************************************************
                        Ovladani virtualni obrazovky
*****************************************************************************)

procedure _SetDisplayOrigin(x,y:word); assembler;
Asm
mov AX,$4F07
xor BX,BX
mov CX,x
mov DX,y
int $10
End;{_setdisplayorigin}

procedure _SetAP(kterou:byte);
Begin
__lhy:=(_maxy+1)*kterou;
End;{_setap}

procedure _SetVP(kterou:byte);
Begin
_setdisplayorigin(0,(_maxy+1)*kterou);
End;{_setvp}

function _PujdeStrankovani:boolean;
Begin
_pujdestrankovani:=(_mode<>0)and((longint(_VelikostVideopameti)shl 16)div((longint(_maxx)+1)*(_maxy+1))>=2);
End;{_pujdestrankovani}

procedure _NastavStrankovani;
Begin
__lhy:=__lhy xor (_maxy+1); {visual page zustava, active se prepne}
End;{_nastavstrankovani}

procedure _ResetStrankovani;
Begin
_setap(0);
_setvp(0);
End;{resetstrankovani}

procedure _swap; assembler;
Asm
xor BX,BX
xor CX,CX     {pro hybani obrazovkou}
mov DX,$03DA  {pro cekani na paprsek}

mov AX,__lhy
or AX,AX
jz @ActiveNa1

 {visual na 1:}
 @cekani1:
 in AL,DX
 test AL,8               {cekani na paprsek}
 jz @cekani1

 mov AX,$4F07
 mov DX,__lhy             {obrazovka dolu}
 int $10

 {active na 0:}
 xor AX,AX
 mov __lhy,AX             {__lhy nahoru}

 jmp @prepnuto          {hotovo}

@ActiveNa1:

 @cekani2:
 in AL,DX
 test AL,8                {cekani na paprsek}
 jz @cekani2

 {visual na 0:}
 mov AX,$4F07
 xor DX,DX                {obrazovka nahoru}
 int $10

 {active na 1:}
 mov AX,_maxy
 inc AX
 mov __lhy,AX              {__lhy dolu}

@prepnuto:
End;{_swap}

BEGIN
_initVESA;
__font:=nil; __vyska:=0;
__velikostX:=1; __velikostY:=1; __roztec:=8;
__lhy:=0;
END.

{format FNT souboru:

 vyska pisma (byte) - pocet radku (bytu) na jeden znak
 doporucena roztec (byte)
 doporucena velikost (byte) - dolni 4 bity urcuji velikostX, horni velikostY
 data (array[0..255,1..vyska] of byte)

Kazdy byte pole dat je jeden radek pismena. Razeny jsou odshora dolu, od
nulteho do 255. znaku. Nejvyssi bit techto bytu znamena pixel, ktery je
na radku nejvic vlevo.
}
